import itertools
import logging

import numpy
import pandas

from pytolemaic.utils.dmd import DMD

try:
    from pytolemaic.analysis_logic.model_analysis.sensitivity.sensitivity_reports import SensitivityOfFeaturesReport
except:
    logging.warning('issue with import of SensitivityOfFeaturesReport')
    pass

from pytolemaic.utils.base_report import Report
from matplotlib import pyplot as plt

class CovarianceShiftReport(Report):
    def __init__(self, covariance_shift:float, sensitivity_report:SensitivityOfFeaturesReport=None, train:DMD=None, test:DMD=None,
                 medium_lvl=0.7, high_lvl=0.95):
        self.medium_lvl = medium_lvl
        self.high_lvl = high_lvl
        self.covariance_shift = covariance_shift
        self.sensitivity = sensitivity_report

        # train and test data is used for distribution plots
        self.train = train
        self.test = test

    def _features_to_look_at_msg(self):
        if self.sensitivity is None:
            return ''
        else:
            return 'Compare distributions for following features: {}'.format(self.sensitivity.most_important_features())

    def covariance_shift_insights(self):
        insights = []
        covariance_shift = numpy.round(self.covariance_shift, 2)
        if covariance_shift < self.medium_lvl:
            return  insights # ok

        features_to_look_at_msg = self._features_to_look_at_msg()

        if covariance_shift < self.high_lvl:
            insights.append('Covariance shift ={:.2g} is not negligible, there may be some issue with train/test distribution. {}'
                            .format(self.covariance_shift, features_to_look_at_msg))
        else:
            insights.append("Covariance shift ={:.2g} is high! Double check the way you've defined the test set! {}"
                            .format(self.covariance_shift, features_to_look_at_msg))

        return insights

    def to_dict(self, printable=False):
        out = dict(covariance_shift=self.covariance_shift)
        if self.sensitivity is not None:
            out.update(dict(covariance_shift_sensitivity=self.sensitivity.to_dict(printable=printable)))
        return self._printable_dict(out, printable=printable)

    @classmethod
    def to_dict_meaning(cls):
        return dict(covariance_shift = "Measure whether the test and train comes from same distribution. "
                                       "High values (max 1) means the test and train come from different distribution. "
                                       "Low score (min 0) means they come from same distribution (=is ok).",
                    covariance_shift_sensitivity = "Sensitivity report for a model which classify samples to train/test sets. "
                                                   "This report is automatically generated only if covariance shift is high. "
                                                   "To generate it manually, see CovarianceShiftCalculator.")

    def plot(self):
        if self.sensitivity is not None:
            fig, ax = plt.subplots(1,1, figsize=(8,10))
            self.sensitivity.plot_sorted_sensitivities(ax=ax, n_features_to_plot=10)
            ax.set(
                title='Feature Sensitivity for Covariance Shift model',
                xlabel='Sensitivity value')

            if self.train is not None and self.test is not None:
                features_to_look_at = self.sensitivity.most_important_features(n_features=3)
                train, _ = self.train.to_df()
                test, _ = self.test.to_df()

                features_to_look_at = [f for f in features_to_look_at if f in train.columns]
                if len(features_to_look_at) > 0:
                    fig, axs = plt.subplots(len(features_to_look_at), 1, figsize=(10, 10))
                    if len(features_to_look_at)==1:
                        axs = [axs]
                for i, feature in enumerate(features_to_look_at):

                    tmp = pandas.DataFrame({'Distribution in train set' : train[feature],
                                            'Distribution in test set': test[feature]})

                    for col in tmp.columns:
                        axs[i].hist(tmp[col], alpha=0.5, label=col)
                    axs[i].legend()
                    axs[i].set_title('Feature "{}"'.format(feature))
                plt.tight_layout()





    def insights(self):

        return self._add_cls_name_prefix(
            itertools.chain(self.covariance_shift_insights()))







import itertools
import logging
import time

from pytolemaic.analysis_logic.dataset_analysis.dataset_analysis_report import DatasetAnalysisReport
from pytolemaic.analysis_logic.model_analysis.scoring.scoring_report import ScoringFullReport
from pytolemaic.analysis_logic.model_analysis.sensitivity.sensitivity_reports import SensitivityFullReport
from pytolemaic.analysis_logic.quality_report import QualityReport
from pytolemaic.utils.base_report import Report


class PyTrustReport(Report):

    def __init__(self, pytrust):
        self.pytrust = pytrust

    @classmethod
    def _try_catch_report(cls, pytrust, attr):

        ts = time.time()
        try:
            # some report may fail due to lack of available information
            logging.info("Calculating {}...".format(attr))
            out = getattr(pytrust, attr, None)
            logging.info("Calculating {}... Done ({:.1f} seconds)".format(attr, time.time() - ts))
        except:
            logging.info("Failed to calculate {}".format(attr))
            out = None

        return out

    def _reports(self):
        return {key: self._try_catch_report(self.pytrust, key) for key in self.to_dict_meaning().keys()}

    def to_dict(self, printable=False):
        return {key: report.to_dict(printable=printable) for key, report in self._reports().items() if
                report is not None}

    @classmethod
    def to_dict_meaning(cls):
        return {'scoring_report': ScoringFullReport.to_dict_meaning(),
                'quality_report': QualityReport.to_dict_meaning(),
                'dataset_analysis_report': DatasetAnalysisReport.to_dict_meaning(),
                'sensitivity_report': SensitivityFullReport.to_dict_meaning(),
                }

    def plot(self):
        for report in self._reports().values():
            if report is not None:
                report.plot()

    def insights(self) -> list:
        reports = self._reports()
        reports.pop('quality_report')
        l = [report.insights() for report in reports.values() if report is not None]
        if any(l):
            return list(itertools.chain(*l))
        else:
            return []



import unittest

from examples.interesting_examples import adult_example, california_housing_example, kddcup99_example, \
    prediction_uncertainty_adult_example, simple_example_classification, simple_example_regression, \
    prediction_uncertainty_california_example, titanic_example


class TestInterestingExamples(unittest.TestCase):

    def test_simple_example_cla(self):
        simple_example_classification.run()

    def test_simple_example_reg(self):
        simple_example_regression.run()

    # @unittest.skip("adult_dataset example - Takes time")
    def test_adult_dataset(self):
        adult_example.run(fast=True)

    def test_adult_dataset_uncertainty(self):
        prediction_uncertainty_adult_example.run()

    def test_california_dataset_uncertainty(self):
        prediction_uncertainty_california_example.run()

    # @unittest.skip("california_housing_dataset example - Takes time")
    def test_california_housing(self):
        california_housing_example.run(fast=True)

    # @unittest.skip("kddcup99 example - Takes time")
    def test_kddcup99(self):
        kddcup99_example.run(fast=True)

    def test_titanic(self):
        titanic_example.run(fast=True)



from pprint import pprint

from matplotlib import pyplot as plt

## For this example we create train/test data representing a linear function
# both numpy and pandas.DataFrame is ok.
from pytolemaic import DMD
from pytolemaic import Metrics
from pytolemaic import PyTrust
from resources.datasets.linear import LinearClassificationDataset


def run():
    ## For this example we create train/test data representing a linear function
    # PyTrust supports both numpy and pandas.DataFrame.

    # Obtain simple classification dataset. Use LinearRegressionDataset for regression
    dataset = LinearClassificationDataset()
    columns_names = dataset.column_names()

    # for quality report, we need for train/test sets and model
    xtrain, ytrain = dataset.training_data
    xtest, ytest = dataset.get_samples()
    classifier = dataset.get_model()

    ## set metric
    metric = Metrics.recall.name

    ## set splitting strategy
    splitter = 'stratified'

    ## sample meta data (e.g. sample weight) - empty in this example
    sample_meta_train = None
    sample_meta_test = None

    # set the feature names names
    columns_meta = {DMD.FEATURE_NAMES: [name for name in columns_names]}

    pytrust = PyTrust(
        model=classifier,
        xtrain=xtrain, ytrain=ytrain,
        xtest=xtest, ytest=ytest,
        sample_meta_train=sample_meta_train, sample_meta_test=sample_meta_test,
        columns_meta=columns_meta,
        metric=metric,
        splitter=splitter)

    scoring_report = pytrust.scoring_report

    score_value = scoring_report.metric_scores[metric].value
    ci_low = scoring_report.metric_scores[metric].ci_low
    ci_high = scoring_report.metric_scores[metric].ci_high


    print('{} score is {:0.3f}'.format(metric, score_value))
    print('Confidence interval is [{:0.3f}, {:0.3f}]'.format(ci_low, ci_high))

    pprint(scoring_report.to_dict(printable=True), width=160)
    pprint(scoring_report.to_dict_meaning(), width=120)

    scoring_report.plot()


if __name__ == '__main__':
    run()
    plt.show()



import multiprocessing

import numpy
import sklearn
from matplotlib import pyplot as plt
from sklearn.ensemble import RandomForestClassifier, RandomForestRegressor
from sklearn.metrics import brier_score_loss
from sklearn.pipeline import Pipeline

from pytolemaic.utils.constants import REGRESSION, CLASSIFICATION
from pytolemaic.utils.dmd import DMD
from pytolemaic.utils.general import GeneralUtils
from pytolemaic.utils.metrics import Metrics


class UncertaintyModelBase():

    def __init__(self, model, uncertainty_method: str,
                 ptype=None,
                 supported_methods: list = None):
        self.model = model
        self.uncertainty_method = uncertainty_method if uncertainty_method != 'default' else supported_methods[0]
        self.dmd_supported = None
        self.is_classification = GeneralUtils.is_classification(model)
        self.uncertainty_analysis_output = None

        if (self.is_classification and ptype != CLASSIFICATION) or \
                (not self.is_classification and ptype == CLASSIFICATION):
            raise ValueError(
                "{} does not support {}".format(type(self), ptype))

        if self.uncertainty_method not in supported_methods:
            raise NotImplementedError(
                "Uncertainty method {} is not in supported methods={}".format(
                    self.uncertainty_method, supported_methods))

    def uncertainty(self, dmd: DMD):
        raise NotImplementedError("")

    def fit_uncertainty_model(self, dmd_test, **kwargs):
        raise NotImplementedError("")

    def fit(self, dmd_test: DMD, **kwargs):
        self.dmd_supported = GeneralUtils.dmd_supported(model=self.model,
                                                        dmd=dmd_test)

        self.fit_uncertainty_model(dmd_test, **kwargs)
        return self

    def predict(self, dmd: DMD):
        if self.dmd_supported:
            if not isinstance(dmd, DMD):
                dmd = DMD(x=dmd)
            return self.model.predict(dmd)
        else:
            if isinstance(dmd, DMD):
                x = dmd.values
            else:
                x = dmd
            return self.model.predict(x)

    def predict_proba(self, dmd: DMD):
        if self.dmd_supported:
            if not isinstance(dmd, DMD):
                raise ValueError("DMD supported but input is not dmd")
            return self.model.predict_proba(dmd)
        else:
            if isinstance(dmd, DMD):
                x = dmd.values
            else:
                x = dmd
            return self.model.predict_proba(x)

    def plot_calibration_curve(self):
        raise NotImplementedError

    @classmethod
    def _get_decision_paths(cls, values, tree, y, uncertainty_threshold, low):

        if low:
            vals = values[y.ravel() <= uncertainty_threshold]
        else:
            vals = values[y.ravel() >= uncertainty_threshold]

        # vals = vals[:min(100, len(vals)), :]
        # decision_path_nodes = [tree.decision_path(val.reshape(1,-1)).indices for val in vals]
        decision_path_nodes = tree.decision_path(vals)

        return decision_path_nodes

    @classmethod
    def _get_nodes(cls, tree, uncertainty_threshold, low, min_samples_in_node=100):
        # get candidates nodes by n_samples and value.
        ids = tree.tree_.n_node_samples >= min_samples_in_node
        values = tree.tree_.value.ravel()

        impurity = tree.tree_.impurity
        impurity[impurity < 0] = 1e100
        rmse = numpy.sqrt(impurity).ravel()
        rmse[tree.tree_.impurity < 0] = -1

        if low:
            ids = ids & (values <= uncertainty_threshold) & (values >= rmse * 2)
        else:
            ids = ids & (values >= uncertainty_threshold) & (values >= rmse * 2)

        if not any(ids):
            return numpy.array([])

        nodes = numpy.arange(tree.tree_.node_count)[ids]

        # prune nodes
        for i in reversed(nodes):  # high to low
            for children in [tree.tree_.children_left[i], tree.tree_.children_right[i]]:
                nodes[nodes == children] = -1
        nodes = nodes[nodes > 0]
        return nodes

    def uncertainty_analysis(self, dmd_train: DMD = None, dmd_test: DMD = None,
                             n_jobs=multiprocessing.cpu_count() - 1, max_depth=5, min_samples_in_area=0.05,
                             percentile=10) -> dict:
        """
        Provides an analysis of areas with low/high uncertainty. Can assist in understanding where the model is good, and where it's not.
        :param dmd_train:
        :param dmd_test:
        :param n_jobs:
        :param max_depth: Maximum number of conditions to define an area.
        :param min_samples_in_area: Minimal train+test samples in reported area.
        :param percentile: percentile over uncertainty to provide threshold for high and low uncertainties.
        :return:
        """
        if dmd_train is not None:
            dmd = dmd_train
            if dmd_test is not None:
                dmd.append(dmd_test)
        elif dmd_test is not None:
            dmd = dmd_test
        else:
            raise ValueError("Require either dmd_train or dmd_test or both")

        if min_samples_in_area < 1:
            min_samples_in_area = int(dmd.n_samples * min_samples_in_area)

        # train RF on uncertainty output (because it's not necessarity RF in uncertainty estimator)
        estimator = RandomForestRegressor(random_state=0, n_jobs=n_jobs, max_depth=max_depth, n_estimators=50,
                                          min_samples_split=max(min_samples_in_area//2, 2),
                                          max_features="sqrt")

        pipeline = GeneralUtils.simple_imputation_pipeline(
            estimator)  # todo better imputation scheme whenever a model is created by package
        y = self.uncertainty(dmd)
        pipeline.fit(dmd.values, y.ravel())

        # extract RF from pipeline
        names, estimators = zip(*pipeline.steps)
        imputer = estimators[-2]
        estimator = estimators[-1]

        low_uncertainty = numpy.percentile(y, percentile)
        high_uncertainty = numpy.percentile(y, 100 - percentile)

        areas_of_extreme_uncertainty = {'low': [], 'high': []}
        for mode in ['low', 'high']:
            extreme_areas = []
            for tree in estimator.estimators_:
                uncertainty_threshold = low_uncertainty if mode == 'low' else high_uncertainty

                # get nodes of interest
                nodes = self._get_nodes(tree, uncertainty_threshold=uncertainty_threshold, low=mode == 'low',
                                        min_samples_in_node=min_samples_in_area)
                if len(nodes) == 0:
                    continue

                # get relevant decision paths of nodes
                node_paths = self._get_decision_paths(imputer.transform(dmd.values), tree, y,
                                                      uncertainty_threshold=uncertainty_threshold, low=mode == 'low')

                # for each node, get msg
                for node in nodes:
                    # search for appropriate path.
                    found = None
                    for node_path_csr in node_paths:
                        nodes_path = node_path_csr.indices
                        if node in nodes_path:
                            found = nodes_path
                            break

                    assert found is not None

                    # this can be moved elsewhere if needed
                    from pytolemaic.analysis_logic.prediction_analysis.decision_tree_report import DecisionTreeExplainer
                    conditions = DecisionTreeExplainer._list_of_conditions(decision_tree=tree,
                                                                           nodes=found,
                                                                           feature_names=dmd.feature_names)

                    mean = tree.tree_.value[node][0][0]
                    rmse = numpy.sqrt(tree.tree_.impurity[node])

                    msg = "Uncertainty( {} ) = {}.".format(" & ".join(conditions),
                                                           "{:.2g}+-{:.2g}".format(mean, rmse))
                    # extreme_areas.append(msg)
                    extreme_areas.append(dict(where=" & ".join(conditions),
                                              uncertainty="{:.2g}+-{:.1g}".format(mean, rmse),
                                              n_samples="{} out of {}".format(tree.tree_.n_node_samples[node],
                                                                              dmd.n_samples)
                                              ))

            areas_of_extreme_uncertainty[mode] = extreme_areas

        areas_of_extreme_uncertainty = {'Low uncertainty': areas_of_extreme_uncertainty['low'],
                                        'High uncertainty': areas_of_extreme_uncertainty['high']}
        # from pprint import pprint
        # pprint(areas_of_extreme_uncertainty)
        return areas_of_extreme_uncertainty

class UncertaintyModelRegressor(UncertaintyModelBase):

    def __init__(self, model, uncertainty_method='mae'):
        super(UncertaintyModelRegressor, self).__init__(
            model=model, uncertainty_method=uncertainty_method,
            ptype=REGRESSION, supported_methods=['mae', 'rmse', 'quantile'])
        self._n_bins = 10
        self.actual_error = None
        self.mean_predicted_error = None
        self._cal_curve_uncertainty = None

    def fit_uncertainty_model(self, dmd_test, n_jobs=multiprocessing.cpu_count() - 1,
                              metric=Metrics.r2, do_analysis=True, **kwargs):

        dmd_test, cal_curve_samples = dmd_test.split(ratio=0.1)

        if self.uncertainty_method in ['mae', 'default']:
            estimator = RandomForestRegressor(
                random_state=0, n_jobs=n_jobs,
                n_estimators=kwargs.pop('n_estimators', 100))

            self.uncertainty_model = GeneralUtils.simple_imputation_pipeline(
                estimator)

            yp = self.predict(dmd_test)
            self.uncertainty_model.fit(dmd_test.values,
                                       numpy.abs(
                                           dmd_test.target.ravel() - yp.ravel()))
        elif self.uncertainty_method in ['rmse']:
            estimator = RandomForestRegressor(
                random_state=0, n_jobs=n_jobs,
                n_estimators=kwargs.pop('n_estimators', 100))

            self.uncertainty_model = GeneralUtils.simple_imputation_pipeline(
                estimator)

            yp = self.predict(dmd_test)
            self.uncertainty_model.fit(dmd_test.values,
                                       (dmd_test.target.ravel() - yp.ravel()) ** 2)
        elif self.uncertainty_method in ['quantile']:
            if isinstance(self.model, Pipeline):

                names, estimators = zip(*self.model.steps)
                estimator = estimators[-1]
                if hasattr(estimator, 'estimators_'):
                    self.model.estimators_ = estimator.estimators_

            if not hasattr(self.model, 'estimators_'):
                raise ValueError(
                    "'quantile' method can work only for models with estimators_ attrbiute. Use method='mae' instead")

        else:
            raise NotImplementedError("Method {} is not implemented"
                                      .format(self.uncertainty_method))

        if do_analysis:
            self.uncertainty_analysis_output = self.uncertainty_analysis(dmd_train=None, dmd_test=dmd_test)

            # the following section is purely for plotting/analysis purposes

            ## calibration curve
            y_pred = self.predict(cal_curve_samples).ravel()
            y_true = cal_curve_samples.target.ravel()

            delta = numpy.abs(y_true - y_pred)
            uncertainty = self.uncertainty(cal_curve_samples).ravel()
            self._cal_curve_uncertainty = uncertainty

            bins = numpy.linspace(0., max(uncertainty) + 1e-8, self._n_bins + 1).ravel()
            binids = numpy.digitize(uncertainty, bins) - 1

            bin_sums = numpy.bincount(binids, weights=uncertainty, minlength=len(bins))
            bin_true = numpy.bincount(binids, weights=delta, minlength=len(bins))
            bin_total = numpy.bincount(binids, minlength=len(bins))

            nonzero = bin_total > 10
            self.actual_error = (bin_true[nonzero] / bin_total[nonzero])
            self.mean_predicted_error = (bin_sums[nonzero] / bin_total[nonzero])

            # calibration curve by metric

            performance = []
            uncertainty_levels_middle = []
            for ibin in range(len(bins) - 1):
                inds = binids == ibin
                if numpy.sum(inds) < 10:
                    continue

                subset_score = metric.function(y_true=y_true[inds], y_pred=y_pred[inds])
                performance.append(subset_score)
                uncertainty_levels_middle.append((bins[ibin] + bins[ibin + 1]) / 2)

            self._cal_curve_metric = {'uncertainty': uncertainty_levels_middle,
                                      'score': performance,
                                      'metric': metric.name}

    def uncertainty(self, dmd: DMD):
        if isinstance(dmd, DMD):
            x = dmd.values
        else:
            x = dmd

        if self.uncertainty_method in ['mae']:
            out = self.uncertainty_model.predict(x)
            return out.reshape(-1, 1)
        elif self.uncertainty_method in ['rmse']:
            out = numpy.sqrt(self.uncertainty_model.predict(x))
            return out.reshape(-1, 1)
        elif self.uncertainty_method in ['quantile']:

            percentile = 10

            ys = numpy.zeros((len(x), len(self.model.estimators_)))
            for i, pred in enumerate(self.model.estimators_):
                ys[:, i] = pred.predict(x)

            err_down = numpy.percentile(ys, percentile, axis=1)
            err_up = numpy.percentile(ys, 100 - percentile, axis=1)

            return ((err_up - err_down) / 2).reshape(-1, 1)

        else:
            raise NotImplementedError("Method {} is not implemented"
                                      .format(self.uncertainty_method))

    def plot_calibration_curve(self):
        fig = plt.figure(figsize=(10, 10))
        ax1 = plt.subplot2grid((3, 1), (0, 0), rowspan=2)
        ax2 = plt.subplot2grid((3, 1), (2, 0))

        ax1.plot([min(self.mean_predicted_error), max(self.mean_predicted_error)],
                 [min(self.mean_predicted_error), max(self.mean_predicted_error)], "k:", label="Perfectly calibrated")

        ax1.plot(self.mean_predicted_error, self.actual_error, "s-",
                 color='b')

        # todo: remove _cal_curve_uncertainty from self
        ax2.hist(self._cal_curve_uncertainty, range=(0, 1.),
                 bins=self._n_bins,
                 color='b',
                 histtype="step", lw=2)

        ax1.set_ylabel("Actual error")
        ax1.set_ylim([min(self.actual_error) * 0.95, max(self.actual_error) * 1.05])
        ax1.legend(loc="lower right")
        ax1.set_title("Calibartion Curve for method {} (Regression)".format(self.uncertainty_method))

        ax2.set_xlabel("Mean predicted error")
        ax2.set_ylabel("Count")
        ax2.legend(loc="upper center", ncol=2)

        # curve by metric
        plt.figure()
        plt.plot(self._cal_curve_metric['uncertainty'], self._cal_curve_metric['score'], '*-r')

        plt.xlabel("Uncertainty level ({})".format(self.uncertainty_method))
        plt.ylabel("{} score".format(self._cal_curve_metric['metric']))
        plt.title("{} score vs uncertainty level".format(self._cal_curve_metric['metric']))



class UncertaintyModelClassifier(UncertaintyModelBase):

    def __init__(self, model, uncertainty_method='confidence'):
        super(UncertaintyModelClassifier, self).__init__(model=model,
                                                         uncertainty_method=uncertainty_method,
                                                         ptype=CLASSIFICATION,
                                                         supported_methods=['confidence', 'probability'])

        self._brier_loss = -1
        self._fraction_of_positives = None
        self._mean_uncertainty = None
        self._cal_curve_uncertainty = None
        self._n_bins = 10

    def fit_uncertainty_model(self, dmd_test, n_jobs=multiprocessing.cpu_count() - 1,
                              metric=Metrics.recall, do_analysis=True,
                              **kwargs):

        if self.uncertainty_method in ['probability']:
            cal_curve_samples = dmd_test
            # no fit logic required

        elif self.uncertainty_method in ['confidence']:
            if do_analysis:
                dmd_test, cal_curve_samples = dmd_test.split(ratio=0.1)

            estimator = RandomForestClassifier(
                random_state=0, n_jobs=n_jobs, n_estimators=100)

            self.uncertainty_model = GeneralUtils.simple_imputation_pipeline(
                estimator)

            y_pred = self.predict(dmd_test)
            is_correct = numpy.array(y_pred.ravel() == dmd_test.target.ravel(),
                                     dtype=int)

            # bug here
            self.uncertainty_model.fit(dmd_test.values, is_correct.ravel())

        else:
            raise NotImplementedError("Method {} is not implemented"
                                      .format(self.uncertainty_method))

        if do_analysis:
            self.uncertainty_analysis_output = self.uncertainty_analysis(dmd_train=None, dmd_test=dmd_test)

            # analysis for plots :

            # calibration curve

            y_pred = self.predict(cal_curve_samples).ravel()
            y_true = cal_curve_samples.target.ravel()
            uncertainty = self.uncertainty(cal_curve_samples)
            self._cal_curve_uncertainty = uncertainty

            self._fraction_of_positives, self._mean_uncertainty = sklearn.calibration.calibration_curve(
                y_true=y_pred == y_true,
                y_prob=uncertainty,
                normalize=True,
                n_bins=self._n_bins,
                strategy='uniform')

            sample_weight = None
            self._brier_loss = brier_score_loss(
                y_true=y_pred == y_true,
                y_prob=1 - uncertainty,
                sample_weight=sample_weight,
                pos_label=1)

            # calibration curve by metric

            uncertainty = uncertainty / max(uncertainty)

            bins = numpy.linspace(0., max(uncertainty) + 1e-8, 5 + 1)
            binids = numpy.digitize(uncertainty.flatten(), bins.ravel()) - 1

            performance = []
            uncertainty_levels_middle = []
            for ibin in range(len(bins) - 1):
                inds = binids == ibin
                if numpy.sum(inds) < 10:
                    continue

                subset_score = metric.function(y_true=y_true[inds], y_pred=y_pred[inds])
                performance.append(subset_score)
                uncertainty_levels_middle.append((bins[ibin] + bins[ibin + 1]) / 2)

            self._cal_curve_metric = {'uncertainty': uncertainty_levels_middle,
                                  'score': performance,
                                  'metric': metric.name}

    def uncertainty(self, dmd: DMD):

        if self.uncertainty_method in ['probability']:
            yproba = self.predict_proba(dmd)
            # break ties
            yproba += 1e-10 * numpy.random.RandomState(0).rand(*yproba.shape)
            max_probability = numpy.max(yproba, axis=1).reshape(-1, 1)
            delta = max_probability - yproba
            yproba[delta == 0] = 0

            # delta[numpy.sum(delta, axis=1)>=20,:] = 0 # i

            # ratio of 2nd highest probability to max_probability
            out = numpy.max(yproba, axis=1).reshape(-1, 1) / max_probability
            return GeneralUtils.f5(out).reshape(-1, 1)
        elif self.uncertainty_method in ['confidence']:
            if isinstance(dmd, DMD):
                x = dmd.values
            else:
                x = dmd
            # return the probability it's a mistake
            out = self.uncertainty_model.predict_proba(x)[:, 0]
            return GeneralUtils.f5(out).reshape(-1, 1)
        else:
            raise NotImplementedError("Method {} is not implemented"
                                      .format(self.uncertainty_method))

    def plot_calibration_curve(self):
        fig = plt.figure(figsize=(10, 10))
        ax1 = plt.subplot2grid((3, 1), (0, 0), rowspan=2)
        ax2 = plt.subplot2grid((3, 1), (2, 0))

        ax1.plot([0, 1], [1, 0.5], "k:", label="Perfectly calibrated")

        ax1.plot(self._mean_uncertainty, self._fraction_of_positives, "s-",
                 color='b',
                 label="brier loss=%1.3f" % self._brier_loss)

        # todo: remove y_proba from self
        ax2.hist(self._cal_curve_uncertainty, range=(0, 1.),
                 bins=self._n_bins,
                 color='b',
                 histtype="step", lw=2)

        ax1.set_ylabel("Fraction of correct predictions")
        ax1.set_ylim([-0.05, 1.05])
        ax1.legend(loc="lower right")
        ax1.set_title("Calibartion Curve for method {}".format(self.uncertainty_method))

        ax2.set_xlabel("Mean uncertainty")
        ax2.set_ylabel("Count")
        ax2.legend(loc="upper center", ncol=2)

        # curve by metric
        plt.figure()
        plt.plot(self._cal_curve_metric['uncertainty'], self._cal_curve_metric['score'], '*-r')

        plt.xlabel("Uncertainty level ({})".format(self.uncertainty_method))
        plt.ylabel("{} score".format(self._cal_curve_metric['metric']))
        plt.title("{} score vs uncertainty level".format(self._cal_curve_metric['metric']))



import itertools

import numpy
from matplotlib import pyplot as plt

from pytolemaic.analysis_logic.dataset_analysis.covariance_shift_report import CovarianceShiftReport
from pytolemaic.utils.base_report import Report


class MissingValuesReport(Report):
    def __init__(self, nan_counts_features, nan_counts_samples, n_samples_to_show=10):
        self._nan_counts_features = nan_counts_features
        self._nan_counts_samples = nan_counts_samples
        self._n_samples_to_show = n_samples_to_show

    @property
    def nan_counts_features(self):
        return self._nan_counts_features

    @property
    def nan_counts_samples(self):
        return self._nan_counts_samples

    def to_dict(self, printable=False):
        out = dict(per_feature={}, per_sample={})
        for nan_ratio, values in self.nan_counts_features.items():
            out['per_feature']['over {}%'.format(nan_ratio * 100)] = dict(count=len(values),
                                                                          threshold=nan_ratio,
                                                                          features=values)

        for nan_ratio, values in self.nan_counts_samples.items():
            if printable:
                n_samples_to_show = min(self._n_samples_to_show, len(values))
                values_in_dict = dict(
                    sorted(values.items(), key=lambda kv: kv[1], reverse=True)[:n_samples_to_show])
            else:
                values_in_dict = values

            out['per_sample']['over {}%'.format(nan_ratio * 100)] = dict(count=len(values),
                                                                         threshold=nan_ratio,
                                                                         samples=values_in_dict)

        # out['per_feature'] = {'scheme': {'threshold': {'count': 'int', 'features': {'feature': 'ratio'}}},
        #                       'values': out['per_feature']}
        # out['per_sample'] = {'scheme': '{threshold: {count: int, sampless: {samples_index: ratio}}} }',
        #                      'values': out['per_sample']}
        return self._printable_dict(out, printable=printable)

    @classmethod
    def to_dict_meaning(cls):
        return dict(per_feature={'over xx%': {'count': 'Number of features with more than xx% missing values',
                                              'threshold': 'threshold for missing values ratio',
                                              'features': 'dict of the form {feature_name : ratio_of_missing_values}.'}},
                    per_sample={'over xx%': {'count': 'Number of samples with more than xx% missing values',
                                             'threshold': 'threshold for missing values ratio',
                                             'samples': 'dict of the form {sample_index : ratio_of_missing_values}.'}})

    def plot(self, n_features_to_plot=10):
        as_str = lambda k: [str(s) for s in k]
        need_plot_for_features = any([len(v) > 0 for v in self.nan_counts_features.values()])
        need_plot_for_samples = any([len(v) > 0 for v in self.nan_counts_samples.values()])
        n_plots = int(need_plot_for_features) + int(need_plot_for_samples)
        if n_plots == 0:
            return

        # list of tuples
        f11 = sorted(self.nan_counts_features.items(), key=lambda kv: kv[0])  # sort by nan_ratio
        f21 = sorted(self.nan_counts_samples.items(), key=lambda kv: kv[0])  # sort by nan_ratio

        # list of tuples
        f12 = sorted(self.nan_counts_features[f11[0][0]].items(), key=lambda kv: kv[1], reverse=True)
        f22 = sorted(self.nan_counts_samples[f21[0][0]].items(), key=lambda kv: kv[1], reverse=True)
        f12 = f12[:min(len(f12), n_features_to_plot)]
        f22 = f22[:min(len(f22), n_features_to_plot)]

        if n_plots == 1:
            fig1, axs = plt.subplots(1, 2, figsize=(10, 5))

            ax11, ax12, ax21, ax22 = [None] * 4
            if need_plot_for_features:
                ax11, ax12 = axs
            else:
                ax21, ax22 = axs
        else:
            fig1, (ax11, ax12) = plt.subplots(1, 2, figsize=(10, 5))
            fig2, (ax21, ax22) = plt.subplots(1, 2, figsize=(10, 5))

        if ax11 is not None:
            keys, values = zip(*f11)
            x = list(reversed(keys))
            y = list(reversed([len(v) for v in values]))
            ax11.bar(x, y, width=0.2)
            ax11.set(
                title='Features with missing values above threshold',
                xlabel='Missing values threshold ',
                ylabel='Number of features',
                xticks=x)

            keys, values = zip(*f12)
            ax12.barh(as_str(reversed(keys)), list(reversed(values)))
            ax12.set(
                title='Features with highest missing ratio',
                xlabel='Missing values ratio',
                yticks=as_str(reversed(keys)))

        if ax22 is not None:
            keys, values = zip(*f21)
            x = list(reversed(keys))
            y = list(reversed([len(v) for v in values]))
            ax21.bar(x, y, width=0.2)
            ax21.set(
                title='Samples with missing values above threshold',
                xlabel='Missing values threshold ',
                ylabel='Number of samples',
                xticks=x)

            keys, values = zip(*f22)
            ax22.barh(as_str(reversed(keys)), list(reversed(values)))
            ax22.set(
                title='Samples with highest missing ratio',
                xlabel='Missing values ratio',
                yticks=as_str(reversed(keys)))

        plt.tight_layout()
        plt.draw()

    def insights(self):
        insights = []
        info = self.to_dict(printable=True)

        lvl1 = 0.1
        lvl2 = 0.5
        lvl3 = 0.9

        counts = [info['per_feature'][key]['count'] for key in info['per_feature']]
        if len(counts) > 0 and max(counts) > 0:
            max_key_features = max([key for key, values in info['per_feature'].items() if values['count'] > 0])
            missing_values = info['per_feature'][max_key_features]
            threshold = missing_values['threshold']

            sentence = "missing values ({}) were found in {} features".format(max_key_features, missing_values['count'])
            if threshold < lvl1:
                insights.append("Some {}.".format(sentence))
            elif threshold < lvl2:
                insights.append("Significant number of {}.".format(sentence))
            elif threshold < lvl3:
                insights.append(
                    "High number of {}. Check out these features:\n{}".format(sentence, missing_values['features']))
            else:
                insights.append("Extremely high number of {}. Check out these features:\n{}".format(sentence,
                                                                                                    missing_values[
                                                                                                        'features']))

        lvl1 = 0.1
        lvl2 = 0.5
        lvl3 = 0.9

        counts = [info['per_sample'][key]['count'] for key in info['per_sample']]
        if len(counts) > 0 and max(counts) > 0:
            max_key_samples = max([key for key, values in info['per_sample'].items() if values['count'] > 0])
            missing_values = info['per_sample'][max_key_samples]
            threshold = missing_values['threshold']

            sentence = "missing values ({}) were found in {} samples".format(max_key_samples, missing_values['count'])
            if threshold < lvl1:
                insights.append("Some {}.".format(sentence))
            elif threshold < lvl2:
                insights.append("Significant number of {}.".format(sentence))
            elif threshold < lvl3:
                insights.append(
                    "High number of {}. Check out these samples:\n{}".format(sentence, missing_values['samples']))
            else:
                insights.append("Extremely high number of {}. Check out these samples:\n{}".format(sentence,
                                                                                                   missing_values[
                                                                                                       'samples']))

        return self._add_cls_name_prefix(insights)


class DatasetAnalysisReport(Report):
    def __init__(self, class_counts, outliers_count, missing_values_report: MissingValuesReport,
                 covariance_shift_report:CovarianceShiftReport=None):
        self._class_counts = class_counts
        self._outliers_count = outliers_count
        self._missing_values_report = missing_values_report
        self._covariance_shift_report = covariance_shift_report


    @property
    def class_counts(self) -> dict:
        return self._class_counts

    @property
    def outliers_count(self) -> dict:
        return self._outliers_count

    @property
    def missing_values_report(self)->MissingValuesReport:
        return self._missing_values_report

    @property
    def covariance_shift_report(self)->CovarianceShiftReport:
        return self._covariance_shift_report

    def to_dict(self, printable=False):
        out = dict(few_class_representatives=self.class_counts,
                   outliers_count=self.outliers_count,
                   missing_values=self.missing_values_report.to_dict(printable=printable),
                   covariance_shift=None,
                   )
        if self.covariance_shift_report is not None:
            out.update(dict(covariance_shift=self.covariance_shift_report.to_dict(printable=printable)))
        return self._printable_dict(out, printable=printable)

    @classmethod
    def to_dict_meaning(cls):
        return dict(
            few_class_representatives='{feature_name: {class: instances_count}} : listing categorical features that has at least 1 class which is under-represented (less than 10 instances).',
            outliers_count='{feature_name: {n-sigma: outliers_info}} : listing numerical features that has more outliers than is expected with respect to n-sigma. '
                           'E.g. for 3-sigma we expect ceil(0.0027 x n_samples) outliers.',
            missing_values=MissingValuesReport.to_dict_meaning(),
            covariance_shift=CovarianceShiftReport.to_dict_meaning(),
        )

    def _plot_class_counts(self):
        if len(self.class_counts) == 0:
            return

        plt.figure()
        ax = plt.subplot()
        features = sorted(self.class_counts.keys())
        n_low_count_classes = [len(self.class_counts[feature]) for feature in features]

        x = [str(k) for k in reversed(features)]
        y = list(reversed(n_low_count_classes))
        plt.barh(x, y)
        ax.set(
            title='Features with under represented classes',
            xlabel='Number of under represented classes',
            ylabel='Features',
            yticks=x)

    def _plot_outlier_counts(self):
        if len(self.outliers_count) == 0:
            return

        plt.figure()
        ax = plt.subplot()
        sigmas = set(list(itertools.chain(*[v.keys() for v in self.outliers_count.values()])))

        features = sorted([str(k) for k in self.outliers_count.keys()], reverse=True)
        for sigma in sorted(sigmas):
            fv = [(feature, values[sigma]['n_outliers']) for feature, values in self.outliers_count.items() if
                  sigma in values]

            x, y = zip(*fv)

            x = [str(k) for k in list(reversed(x))]
            y = list(reversed(y))
            plt.barh(x, y)
        ax.set(
            title='Features with outliers',
            xlabel='Number of outliers',
            ylabel='Features',
            yticks=features)
        ax.legend(sorted(sigmas))

    def plot(self):
        self.missing_values_report.plot()
        self._plot_class_counts()
        self._plot_outlier_counts()
        if self.covariance_shift_report is not None:
            self.covariance_shift_report.plot()

    def _class_count_insights(self):
        insights = []
        for feature_name, class_counts in self.class_counts.items():
            if len(class_counts) == 1:
                sentence = "Feature '{}' contains a class with few representatives".format(feature_name)
            else:
                sentence = "Feature '{}' contains {} classes with few representatives - {} samples in total. " \
                           "With so few samples in each class there would be little to no learning.\n\tConsider merging all classes into a single 'other' class" \
                    .format(feature_name, len(class_counts), sum(class_counts.values()))

            list_of_classes = ["class '{}' has only {} representatives".format(key, value) for key, value in
                               class_counts.items()]
            if len(class_counts) == 1:
                list_of_classes = ": " + list_of_classes[0]
            else:
                n_lines_to_show = 5
                if len(list_of_classes) > n_lines_to_show:
                    suffix = '\n\t... {} more.'.format(len(list_of_classes) - n_lines_to_show)
                else:
                    suffix = ''
                list_of_classes = ". List of classes:\n\t" + ",\n\t".join(list_of_classes[:n_lines_to_show]) + suffix

            insights.append("{}{}".format(sentence, list_of_classes))

        return self._add_cls_name_prefix(insights)

    def _outlier_counts_insights(self):
        insights = []
        for feature_name, outliers_info in self.outliers_count.items():
            max_n_sigma = max(outliers_info.keys())

            n_sigma_info = outliers_info[max_n_sigma]
            n_sigma_range = "[{:.3g},{:.3g}]".format(
                n_sigma_info['mean'] - n_sigma_info['n_sigma'] * n_sigma_info['std'],
                n_sigma_info['mean'] + n_sigma_info['n_sigma'] * n_sigma_info['std'], )
            sentence = "Feature '{}' has {} outliers with respect to {} - a {} range." \
                .format(feature_name, n_sigma_info['n_outliers'], n_sigma_range, max_n_sigma)
            insights.append(sentence)

            log10_plus_one = numpy.log10(numpy.round(n_sigma_info['max'] + 1))

            if n_sigma_info['max'] >= 99 and log10_plus_one == int(log10_plus_one):
                insights.append("Feature '{}' has a maximal value of {} which doesn't seem to be a legit value".format(
                    feature_name, n_sigma_info['max']))

        return self._add_cls_name_prefix(insights)

    def insights(self):
        return list(itertools.chain(self.missing_values_report.insights(),
                                    self.covariance_shift_report.insights(),
                                    self._outlier_counts_insights(),
                                    self._class_count_insights()))



from pprint import pprint

from pytolemaic import Metrics
from pytolemaic import PyTrust
from pytolemaic.utils.general import GeneralUtils, tic, toc
from resources.datasets.uci_adult import UCIAdult


def run(fast=False):
    dataset = UCIAdult()
    classifier = dataset.get_model()
    train, test = dataset.as_dmd()

    metric = Metrics.recall.name

    pytrust = PyTrust(
        model=classifier,
        xtrain=train,
        xtest=test,
        metric=metric)

    print("We've trained a ML model (details below) on uci adult dataset. Let's see whether our model is a good one")
    print("Model details\n", classifier, '\n\n')

    print("Let's analyze the dataset")
    tic("dataset_analysis_report")
    dataset_analysis_report = pytrust.dataset_analysis_report
    pprint(dataset_analysis_report.to_dict(printable=True))
    print('\n'.join(dataset_analysis_report.insights()))
    toc("dataset_analysis_report")

    dataset_analysis_report.plot()

    print("Let's calculate score report")
    tic("scoring_report")
    scoring_report = pytrust.scoring_report
    toc("scoring_report")

    print("\nNow let's deepdive into the report!")
    scoring_report_deepdive(scoring_report)

    print("\n\nNext we'd like to check feature sensitivity")
    tic("sensitivity_report")
    sensitivity_report = pytrust.sensitivity_report
    toc("sensitivity_report")

    print("\nNow let's deepdive into the report!")
    sensitivity_deepdive(sensitivity_report)

    print("\nFinally let's review overall quality score!")
    quality_report = pytrust.quality_report

    print("Overall quality of train data: {:0.3f}".format(quality_report.train_quality_report.train_set_quality))
    print("Overall quality of test data: {:0.3f}".format(quality_report.test_quality_report.test_set_quality))
    print("Overall quality of model: {:0.3f}".format(quality_report.model_quality_report.model_quality))
    print('*** quality_report was commented out ***')
    # pprint(quality_report.to_dict(printable=True), width=120)
    # pprint(quality_report.to_dict_meaning(), width=120)

    print("Let's check for insights...")
    tic("insights_summary")
    print('\n'.join(pytrust.insights))
    toc("insights_summary")

    print("\nLet's create a Lime explainer")
    lime_explainer = pytrust.create_lime_explainer(max_samples=6000 if fast else 64000)

    sample = test.values[0, :]
    print("And plot explanation for the first sample in test data: {}".format(sample))
    lime_explainer.plot(sample)
    explanation = lime_explainer.explain(sample)
    print("Lime explanation is:")
    pprint(GeneralUtils.round_values(explanation))

    #pprint(quality_report.to_dict(printable=True), width=120)
    # pprint(quality_report.to_dict_meaning(), width=120)


def sensitivity_deepdive(sensitivity_report):
    print("\nlet's check which 3 features are most important. Does it make sense?")
    print(sensitivity_report.shuffle_report.sorted_sensitivities[:3])
    print("Looking on top 3 features in sensitivity to missing values we see there is some difference")
    print(sensitivity_report.missing_report.sorted_sensitivities[:3])
    print("This means that error caused by missing values affect the model differently than a regular mistake")
    print("\nThere are {} features with 0 sensitivity, and {} features with low sensitivity".format(
        sensitivity_report.shuffle_report.stats_report.n_zero, sensitivity_report.shuffle_report.stats_report.n_low))
    print("\nlet's check which 3 features are least important. Does it make sense?")
    print(sensitivity_report.shuffle_report.sorted_sensitivities[-3:])

    print(
        "\nUsing the sensitivity report we can obtain some vulnerability measures (lower is better). The meaning of the fields can be obtained with to_dict_meaning()")
    pprint(sensitivity_report.vulnerability_report.to_dict(printable=True), width=120)
    print('*** vulnerability_report explanation was commented out ***')
    # pprint(sensitivity_report.vulnerability_report.to_dict_meaning(), width=120)

    print("We see that the imputation measure is relatively high, which means the model is sensitive "
          "to imputation method")
    print("However, none of the values seems to be high, which is reassuring")

    print("\n\nWe can see entire sensitivity report as well as explanation for the various fields using"
          " to_dict() and to_dict_meaning()")
    print('*** sensitivity_report was commented out ***')
    # pprint(sensitivity_report.to_dict(printable=True), width=120)
    # pprint(sensitivity_report.to_dict_meaning(), width=120)
    print("\nNow let's plot some nice graphs")
    sensitivity_report.plot()


def scoring_report_deepdive(scoring_report):
    metric = scoring_report.target_metric
    score_value = scoring_report.metric_scores[metric].value
    ci_low = scoring_report.metric_scores[metric].ci_low
    ci_high = scoring_report.metric_scores[metric].ci_high
    ci_ratio = scoring_report.metric_scores[metric].ci_ratio

    print("\nLet's check the target score first - ")
    print('{} score is {:0.3f} which is reasonable'.format(metric, score_value))
    print('Confidence interval is [{:0.3f}, {:0.3f}] which implies a ci_ratio of {:0.3f} which is quite good'.format(
        ci_low, ci_high, ci_ratio))

    print("\nNow check out the confusion matrix - regular and normalized version")
    print('Regular confusion matrix:\n', scoring_report.confusion_matrix.confusion_matrix[0], '\n',
          scoring_report.confusion_matrix.confusion_matrix[1], '\n')
    print('Normalized Confusion matrix:\n', scoring_report.confusion_matrix.normalized_confusion_matrix[0], '\n',
          scoring_report.confusion_matrix.normalized_confusion_matrix[1])
    print("Looking on matrices we see that the recall for class {} is good, but the recall on class {} is not.".format(
        scoring_report.confusion_matrix.labels[0], scoring_report.confusion_matrix.labels[1]))


    print("\nFinally, let's look on the separation quality")

    print(
        "\nWe can see entire scoring report as well as explanation for the various fields using to_dict() and to_dict_meaning()")
    print('*** scoring_report was commented out ***')
    # pprint(scoring_report.to_dict(printable=True), width=120)
    # pprint(scoring_report.to_dict_meaning(), width=120)

    print("\nAnd also plot some nice graphs")
    scoring_report.plot()


if __name__ == '__main__':
    from matplotlib import pyplot as plt

    run()
    plt.show()



import itertools

import numpy
import sklearn.calibration
from matplotlib import pyplot as plt
from sklearn.metrics import confusion_matrix, classification_report, \
    RocCurveDisplay, PrecisionRecallDisplay, \
    precision_recall_curve, average_precision_score, auc, roc_curve, \
    brier_score_loss
from sklearn.utils.multiclass import unique_labels

from pytolemaic.utils.base_report import Report
from pytolemaic.utils.general import GeneralUtils
from pytolemaic.utils.metrics import Metrics


class ROCCurveReport(Report):
    def __init__(self, y_true, y_proba, labels=None, sample_weight=None):
        self._labels = labels if labels is not None else unique_labels(y_true).tolist()
        self._roc_curve = {}
        self._auc = {}
        for class_index, label in enumerate(self.labels):
            fpr, tpr, thresholds = roc_curve(y_true == class_index, y_proba[:, class_index],
                                             pos_label=1, sample_weight=sample_weight,
                                             drop_intermediate=True)

            self._roc_curve[label] = dict(fpr=fpr, tpr=tpr, thresholds=thresholds)
            self._auc[label] = auc(fpr, tpr)

    @property
    def labels(self):
        return self._labels

    @property
    def roc_curve(self):
        return self._roc_curve

    @property
    def auc(self):
        return self._auc

    def to_dict(self, printable=False):
        out = dict(roc_curve=self.roc_curve,
                   auc=self.auc,
                   labels=self.labels)
        return self._printable_dict(out, printable=printable)

    @classmethod
    def to_dict_meaning(cls):
        return dict(
            roc_curve="ROC curve as dict (fpr, tpr, and thresholds), per label",
            auc='Area under curve per label',
            labels="The class labels"
        )

    def plot(self, ax=None, figsize=(10,5)):
        if ax is None:
            fig, ax = plt.subplots(1,1,figsize=figsize)

        ax.set_title("ROC Curve")
        possible_colors = GeneralUtils.shuffled_colors()
        for class_index, label in enumerate(self.labels):
            fpr, tpr = self._roc_curve[label]['fpr'], self._roc_curve[label]['tpr']
            roc_auc = self.auc[label]
            viz = RocCurveDisplay(fpr=fpr, tpr=tpr, roc_auc=roc_auc, estimator_name='Classifier')

            viz.plot(ax=ax, name=label, color=possible_colors[class_index])

        plt.draw()

    def insights(self):
        min_points_threshold = 3 + 2  # 3 + '0' + '1'

        thresholds = {label: set(numpy.clip(self._roc_curve[label]['thresholds'], 0, 1)) for label in self.labels}

        insights = []

        for label in self.labels:
            n_points = len(thresholds[label])
            if n_points <= min_points_threshold:
                insights = ["Only {} probability values ({}) for class '{}'. "
                            "This may impede the tuning of prediction threshold and the calibartion curve. Such behavior may indicate a bug."
                                .format(n_points, sorted(thresholds[label]), label)]

        return self._add_cls_name_prefix(insights)


class PrecisionRecallCurveReport(Report):
    def __init__(self, y_true, y_proba, labels=None, sample_weight=None):
        self._labels = labels if labels is not None else unique_labels(y_true).tolist()
        self._recall_precision_curve = {}
        self._average_precision = {}
        for class_index, label in enumerate(self.labels):
            precision, recall, thresholds = precision_recall_curve(y_true == class_index, y_proba[:, class_index],
                                                                   pos_label=1, sample_weight=sample_weight, )

            self._recall_precision_curve[label] = dict(precision=precision, recall=recall, thresholds=thresholds)

            self._average_precision[label] = average_precision_score(y_true == class_index, y_proba[:, class_index],
                                                                     pos_label=1, sample_weight=sample_weight, )

    @property
    def labels(self):
        return self._labels

    @property
    def recall_precision_curve(self):
        return self._recall_precision_curve

    @property
    def average_precision(self):
        return self._average_precision

    def to_dict(self, printable=False):
        out = dict(recall_precision_curve=self.recall_precision_curve,
                   average_precision=self.average_precision,
                   labels=self.labels)
        return self._printable_dict(out, printable=printable)

    @classmethod
    def to_dict_meaning(cls):
        return dict(
            recall_precision_curve="Recall Precision Curve as dict (recall, precision, and thresholds), per label",
            average_precision='Average Precision per label',
            labels="The class labels"
        )

    def plot(self, ax=None, figsize=(10,5)):
        if ax is None:
            fig, ax = plt.subplots(1,1, figsize=figsize)

        ax.set_title("Precision Recall Curve")
        possible_colors = GeneralUtils.shuffled_colors()
        for class_index, label in enumerate(self.labels):
            precision = self._recall_precision_curve[label]['precision']
            recall = self._recall_precision_curve[label]['recall']
            average_precision = self._average_precision[label]

            viz = PrecisionRecallDisplay(precision=precision, recall=recall, average_precision=average_precision,
                                         estimator_name='Classifier')

            viz.plot(ax=ax, name=label, color=possible_colors[class_index])

    def insights(self):
        return []


class CalibrationCurveReport(Report):
    def __init__(self, y_true, y_proba, labels=None, sample_weight=None,
                 n_bins=10):
        self._labels = labels if labels is not None else unique_labels(y_true).tolist()
        self._calibration_curve = {}
        self._brier_loss = {}
        self._y_proba = y_proba
        self._n_bins = n_bins

        for class_index, label in enumerate(self.labels):
            fraction_of_positives, mean_predicted_value = \
                sklearn.calibration.calibration_curve(
                    y_true=y_true == class_index,
                    y_prob=y_proba[:, class_index],
                    normalize=False,
                    n_bins=n_bins,
                    strategy='uniform')
            self._calibration_curve[label] = dict(
                fraction_of_positives=fraction_of_positives,
                mean_predicted_value=mean_predicted_value)
            self._brier_loss[label] = brier_score_loss(
                y_true=y_true == class_index,
                y_prob=y_proba[:, class_index],
                sample_weight=sample_weight,
                pos_label=1)

    @property
    def labels(self):
        return self._labels

    @property
    def calibration_curve(self):
        return self._calibration_curve

    @property
    def brier_loss(self):
        return self._brier_loss

    def to_dict(self, printable=False):
        out = dict(calibration_curve=self.calibration_curve,
                   brier_loss=self.brier_loss,
                   labels=self.labels)
        return self._printable_dict(out, printable=printable)

    @classmethod
    def to_dict_meaning(cls):
        return dict(
            calibration_curve="Calibration curve as a dict (fraction_of_positives, mean_predicted_value), per label",
            brier_loss="Brier loss (lower is better), per label",
            labels="The class labels"
        )

    def plot(self, axs=None, figsize=(10,10)):
        if axs is None:
            fig = plt.figure(figsize=figsize)
            ax1 = plt.subplot2grid((3, 1), (0, 0), rowspan=2)
            ax2 = plt.subplot2grid((3, 1), (2, 0))
        else:
            ax1, ax2 = axs

        ax1.plot([0, 1], [0, 1], "k:", label="Perfectly calibrated")

        ax1.set_title("Calibartion Curve")
        possible_colors = GeneralUtils.shuffled_colors()
        for class_index, label in enumerate(self.labels):
            mean_predicted_value = self._calibration_curve[label][
                'mean_predicted_value']
            fraction_of_positives = self._calibration_curve[label][
                'fraction_of_positives']

            brier_loss = self.brier_loss[label]
            ax1.plot(mean_predicted_value, fraction_of_positives, "s-",
                     color=possible_colors[class_index],
                     label="class %s (brier loss=%1.3f)" % (
                     label, brier_loss))

            # todo: remove y_proba from self
            ax2.hist(self._y_proba[:, class_index], range=(0, 1.),
                     bins=self._n_bins, label=label,
                     color=possible_colors[class_index],
                     histtype="step", lw=2)

        ax1.set_ylabel("Fraction of positives")
        ax1.set_ylim([-0.05, 1.05])
        ax1.legend(loc="lower right")
        ax1.set_title('Calibration plots (reliability curve)')

        ax2.set_xlabel("Mean predicted value")
        ax2.set_ylabel("Count")
        ax2.legend(loc="upper center", ncol=2)

    def insights(self):
        lvl1 = 0.25
        lvl2 = 0.5
        lvl3 = 0.75

        brier_loss = [(label, self._brier_loss[label]) for label in self.labels]
        label, max_loss = sorted(brier_loss, key=lambda pair: pair[1], reverse=True)[0]
        insights = []
        if max_loss <= lvl1:
            pass  # OK
        elif max_loss <= lvl2:
            insights = [
                'Brier loss for class {} indicates model is not well calibrated. Please look at Calibration Curve.'.format(
                    label)]
        elif max_loss <= lvl3:
            insights = [
                'Brier loss for class {} indicates model is badly calibrated! Check the Calibration Curve.'.format(
                    label)]
        else:
            insights = [
                'Brier loss for class {} indicates model is not calibrated at all! Check the Calibration Curve!'.format(
                    label)]

        return self._add_cls_name_prefix(insights)


class SklearnClassificationReport(Report):
    def __init__(self, y_true, y_pred, y_proba, labels=None,
                 sample_weight=None, digits=3):
        self._labels = labels if labels is not None else unique_labels(y_true,
                                                                       y_pred).tolist()

        self._sample_weight = sample_weight

        self._sklearn_performance_summary_text = classification_report(
            y_true=y_true, y_pred=y_pred.reshape(-1, 1),
            labels=None,
            target_names=[str(k) for k in self._labels],
            sample_weight=sample_weight, digits=digits,
            output_dict=False)

        self._sklearn_performance_summary_dict = classification_report(
            y_true=y_true, y_pred=y_pred,
            labels=None, target_names=self._labels,
            sample_weight=sample_weight, digits=digits,
            output_dict=True)

        self._roc_curve = ROCCurveReport(y_true=y_true, y_proba=y_proba,
                                         labels=self.labels,
                                         sample_weight=sample_weight)

        self._calibration_curve = CalibrationCurveReport(y_true=y_true,
                                                         y_proba=y_proba,
                                                         labels=self.labels,
                                                         sample_weight=sample_weight)

        self._precision_recall_curve = PrecisionRecallCurveReport(
            y_true=y_true, y_proba=y_proba,
            labels=self.labels, sample_weight=sample_weight)
        self._y_true = y_true
        self._y_pred = y_pred
        self._y_proba = y_proba

    @property
    def labels(self):
        return self._labels

    @property
    def roc_curve(self) -> ROCCurveReport:
        return self._roc_curve

    @property
    def precision_recall_curve(self) -> PrecisionRecallCurveReport:
        return self._precision_recall_curve

    @property
    def calibration_curve(self) -> CalibrationCurveReport:
        return self._calibration_curve

    @property
    def sklearn_performance_summary(self):
        return self._sklearn_performance_summary_dict

    def to_dict(self, printable=False):
        out = dict(
            sklearn_performance_summary=self.sklearn_performance_summary,
            roc_curve=self.roc_curve.to_dict(printable=printable),
            precision_recall_curve=self.precision_recall_curve.to_dict(printable=printable),
            calibration_curve=self.calibration_curve.to_dict(printable=printable),
            labels=self.labels)
        return self._printable_dict(out, printable=printable)

    @classmethod
    def to_dict_meaning(cls):
        return dict(
            sklearn_performance_summary="Accuracy score for various metrics produced by sklearn",
            roc_curve="ROC curve report",
            precision_recall_curve="Precision-Recall curve report",
            calibration_curve="Calibration curve report",
            labels="The class labels"
        )

    def _plot_classification_report(self):
        import matplotlib.pyplot as plt

        fig = plt.figure(figsize=(12, 1 + len(self.labels)))
        fig.text(0.5, 0.5, self._sklearn_performance_summary_text,
                 ha='center', va='center', size=20, fontname='courier', family='monospace')
        plt.tight_layout()


    def plot(self, figsize=(14,5)):
        self._plot_classification_report()

        fig, (ax1, ax2) = plt.subplots(1, 2, figsize=figsize)

        self.precision_recall_curve.plot(ax1)
        self.roc_curve.plot(ax2)
        plt.tight_layout()

        # new figure
        self.calibration_curve.plot(figsize=figsize)

        plt.tight_layout()
        plt.draw()

    def _sklearn_summary_insights(self):
        lvl1 = 0.75
        lvl2 = 0.5
        lvl3 = 0.25

        insights = []

        metrics = ['f1-score', 'precision', 'recall']

        for label in list(self.labels) + ['macro avg']:
            for metric in metrics:
                score = self.sklearn_performance_summary[label][metric]
                if label == 'macro avg':
                    prefix = 'The overall performance ({}) is '.format(metric)
                else:
                    prefix = '{} score for class {} is {} which is '.format(metric, label, numpy.round(score, 2))

                if score >= lvl1:
                    pass  # ok
                elif score >= lvl2:
                    insights.append(prefix + 'quite low.')
                elif score >= lvl3:
                    insights.append(prefix + 'very low! Look at the confusion matrix.'.format(metric, label))
                elif score >= lvl2:
                    insights.append(prefix + 'extremely low! Check out the confusion matrix!'.format(metric, label))

        return self._add_cls_name_prefix(insights)

    def insights(self):
        return list(itertools.chain(self._sklearn_summary_insights(),
                                    self.roc_curve.insights(),
                                    self.precision_recall_curve.insights(),
                                    self.calibration_curve.insights(),
                                    ))


class ConfusionMatrixReport(Report):
    def __init__(self, y_true, y_pred, labels: list = None):
        self._confusion_matrix = confusion_matrix(y_true=y_true, y_pred=y_pred,
                                                  labels=unique_labels(y_true, y_pred)).tolist()
        self._labels = labels if labels is not None else unique_labels(y_true, y_pred).tolist()
        if isinstance(self._labels, numpy.ndarray):
            self._labels = self._labels.tolist()

    @property
    def labels(self)->list:
        return self._labels

    @property
    def confusion_matrix(self):
        return self._confusion_matrix

    @property
    def normalized_confusion_matrix(self):
        cm = numpy.array(self.confusion_matrix)
        cm = cm / cm.sum(axis=1)[:, numpy.newaxis]

        return GeneralUtils.f3(cm).tolist()

    def to_dict(self, printable=False):
        out = dict(confusion_matrix=self.confusion_matrix,
                    normalized_confusion_matrix=self.normalized_confusion_matrix,
                    labels=self.labels)
        return self._printable_dict(out, printable=printable)

    @classmethod
    def to_dict_meaning(cls):
        return dict(
            confusion_matrix="Confusion Matrix - rows indicate true values, columns indicate predicted value (in contrast to the convention shown in https://en.wikipedia.org/wiki/Confusion_matrix)",
            normalized_confusion_matrix="Normalized confusion matrix - the sum of each rows is equal to 1",
            labels="The class labels"
            )

    @classmethod
    def _plot_confusion_matrix(cls, confusion_matrix, labels: list, title,
                               ax, cmap=plt.cm.Greens):

        cm = numpy.array(confusion_matrix)
        im = ax.imshow(cm, interpolation='nearest', cmap=cmap)

        # ax.figure.colorbar(im, ax=ax)

        ax.set(xticks=[-0.5] + numpy.arange(cm.shape[1]).tolist() + [cm.shape[1] - 0.5],
               yticks=[-0.5] + numpy.arange(cm.shape[0]).tolist() + [cm.shape[0] - 0.5],
               xticklabels=[''] + labels + [''], yticklabels=[''] + labels + [''],
               title=title,
               ylabel='True labels',
               xlabel='Predicted labels')

        # Rotate the tick labels and set their alignment.
        plt.setp(ax.get_xticklabels(), rotation=45, ha="right",
                 rotation_mode="anchor")

        # Loop over data dimensions and create text annotations.
        fmt = '.2g' if numpy.min(cm[cm > 0]) < 1 else 'd'
        if fmt == 'd':
            cm = cm.astype(int)
        # noinspection PyArgumentList
        thresh = cm.max() / 2.
        for i in range(cm.shape[0]):
            for j in range(cm.shape[1]):
                ax.text(j, i, format(cm[i, j], fmt),
                        ha="center", va="center",
                        color="white" if cm[i, j] > thresh else "black")
        return ax

    def plot(self, axs=None, figsize=(12,5)):
        if axs is None:
            fig, axs = plt.subplots(1, 2, figsize=figsize)

        ax1, ax2 = axs

        self._plot_confusion_matrix(confusion_matrix=self.confusion_matrix,
                                    labels=self.labels,
                                    title='Confusion Matrix',
                                    ax=ax1)
        self._plot_confusion_matrix(confusion_matrix=self.normalized_confusion_matrix,
                                    labels=self.labels,
                                    title='Normalized confusion Matrix',
                                    ax=ax2)

        plt.tight_layout()

        # plt.show()

    def insights(self):
        insights = []
        for i, label in enumerate(self.labels):
            if numpy.sum(numpy.array(self.confusion_matrix)[:, i]) < 10:
                insights.append('Model rarely predicts class {}! Is that ok??'.format(label))

        return self._add_cls_name_prefix(insights)

class ScatterReport(Report):
    def __init__(self, y_true, y_pred, error_bars=None):
        self._y_true = numpy.array(y_true).reshape(-1, 1)
        self._y_pred = numpy.array(y_pred).reshape(-1, 1)
        self._error_bars = error_bars

    @property
    def y_true(self):
        return self._y_true

    @property
    def y_pred(self):
        return self._y_pred

    def to_dict(self, printable=False):
        error_bars = self._error_bars.ravel() if self._error_bars is not None else None
        out = dict(y_true=self.y_true.ravel(),
                    y_pred=self.y_pred.ravel(),
                    error_bars=error_bars)
        return self._printable_dict(out, printable=printable)

    @classmethod
    def to_dict_meaning(cls):
        return dict(y_true="True values",
                    y_pred="Predicted values",
                    error_bars="The uncertainty of each prediction")

    def plot(self, max_points=500, figsize=(10,5), residual_plot=True):
        if max_points is None:
            max_points = len(self.y_true)

        rs = numpy.random.RandomState(0)
        inds = rs.permutation(len(self.y_true))[:max_points]

        if residual_plot:
            fig, (ax1, ax2) = plt.subplots(1,2,figsize=figsize)
        else:
            fig, ax1 = plt.subplots(1, 1, figsize=figsize)



        y_true = self.y_true[inds]
        y_pred = self.y_pred[inds]
        error_bars = None if self._error_bars is None else self._error_bars[inds]

        ## scatter plot

        ax = ax1
        ax.errorbar(y_true, y_pred, xerr=None, yerr=error_bars, fmt='.b', ecolor='k')

        mn = numpy.min([y_true.min(), y_pred.min()])
        mx = numpy.max([y_true.max(), y_pred.max()])
        ax.plot([mn, mx], [mn, mx],':k')

        ax.set_xlabel('Y true')
        ax.set_ylabel('Y predicted')
        ax.set_title('Scatter plot')

        ## scatter plot
        if residual_plot:
            ax = ax2
            ax.errorbar(y_true, y_pred-y_true, xerr=None, yerr=error_bars, fmt='.b', ecolor='k')

            mn = numpy.min([y_true.min(), y_pred.min()])
            ax.plot([mn, mx], [0, 0], '-k')

            ax.set_xlabel('Y true')
            ax.set_ylabel('Ypred - Ytrue')
            ax.set_title('Residual plot')

        plt.draw()
        # plt.show()

    def insights(self):
        # todo: what insights can we derive?
        return []

class ScoringMetricReport(Report):
    def __init__(self, metric_name: str, value, ci_low, ci_high):
        self._metric_name = metric_name
        self._value = value
        self._ci_low = ci_low
        self._ci_high = ci_high

    def to_dict(self, printable=False):
        out = dict(
            metric=self.metric_name,
            value=self.value,
            ci_low=self.ci_low,
            ci_high=self.ci_high,
            ci_ratio=self.ci_ratio,
        )
        return self._printable_dict(out, printable=printable)

    @classmethod
    def to_dict_meaning(cls):
        return dict(metric="The metric which was used to calculate the score value",
                    value="The metric value - could be a score value (e.g. auc) or a loss value (e.g. mae), depending on metric",
                    ci_low="Lower confidence interval based on percentile 25",
                    ci_high="Higher confidence interval based on percentile 75",
                    ci_ratio="Measure confidence interval relative size - lower is better. Equation: (ci_high-ci_low)/(ci_low+ci_high)*2",
                    )

    def plot(self, ax=None, figsize=(10,5)):
        if ax is None:
            fig, ax = plt.subplots(1,1,figsize=figsize)


        ci_low = GeneralUtils.f5(self.ci_low)
        ci_high = GeneralUtils.f5(self.ci_high)
        value = GeneralUtils.f5(self.value)
        if ci_high == ci_low:
            n_digits = 5
        else:
            n_digits = -int(numpy.log10(ci_high - ci_low)) + 1  # 0.0011 --> -(-2) +1 = 3

        ax.plot([ci_low, ci_high], [1, 1], '-b',
                ci_low, 1, '|b',
                ci_high, 1, '|b',
                value, 1, 'or', )

        delta = (ci_high - ci_low) * 1e-1 + 10 ** -n_digits / 2

        metric_obj = Metrics.supported_metrics()[self.metric_name]
        metric_full_name = metric_obj.full_name
        r_lim = 1e100 if metric_obj.is_loss else 1
        l_lim = 0 if metric_obj.is_loss else -1e100

        l_lim = max(l_lim, numpy.round(ci_low - delta, n_digits))
        r_lim = min(r_lim, numpy.round(ci_high + delta, n_digits))

        ax.set_xlim(l_lim, r_lim)
        n_points = 1 + int(numpy.round(r_lim - l_lim, n_digits) / 10 ** -n_digits) % 10
        n_points = max(n_points, 3)

        x = numpy.linspace(l_lim, r_lim, num=n_points)

        xlabels = ["%.5g" % numpy.round(k, n_digits) for k in x]
        ax.set(xticks=x.tolist(),
               xticklabels=xlabels,
               yticks=[0.5],
               yticklabels=[''],
               title='Confidence interval for metric {}'.format(metric_full_name),
               ylabel='',
               xlabel='')

        # Loop over data dimensions and create text annotations.
        for x, label in [(ci_low, 'ci_low (25%)'),
                         (value, '{} value: {:.5g}'.format(self.metric_name, numpy.round(value, 1 + n_digits))),
                         (ci_high, 'ci_high  (75%)')]:
            y = 1.01 + 0.01 * (x == value)
            ax.text(x, y, label,
                    ha="center", va="center", )

        plt.draw()

    def insights(self):
        insights = []

        if self.value < self.ci_low or self.value > self.ci_high:
            insights.append('{} value {} is out of range of confidence interval [{},{}]'
                            .format(self.metric_name, self.value, self.ci_low, self.ci_high))

        lvl1 = 0.1
        lvl2 = 0.5
        ci_range_and_ratio = '[{:.3g}, {:.3g}], ci ratio of {:.3g}'.format(self.ci_low, self.ci_high, self.ci_ratio)
        if self.ci_ratio < lvl1:
            pass  # ok
        elif self.ci_ratio < lvl2:
            insights.append(
                'Confidence interval for metric {} is quite large ({})'.format(self.metric_name, ci_range_and_ratio))
        else:
            insights.append(
                'Confidence interval for metric {} is very large ({}). The score measurement of {} is inaccurate.'.format(
                    self.metric_name, ci_range_and_ratio, self.value))

        return self._add_cls_name_prefix(insights)

    @property
    def metric_name(self):
        return self._metric_name

    @property
    def value(self):
        return self._value

    @property
    def ci_low(self):
        return self._ci_low

    @property
    def ci_high(self):
        return self._ci_high

    @property
    def ci_ratio(self):
        # large ci difference is more of a concern if score is high
        ci_low = Metrics.metric_as_loss(value=self.ci_low, metric=self.metric_name)
        ci_high = Metrics.metric_as_loss(value=self.ci_high, metric=self.metric_name)

        if ci_low == ci_high:
            return 0
        else:
            ci_ratio = abs(ci_high - ci_low) / (ci_high + ci_low) * 2
            return ci_ratio


class ScoringFullReport(Report):
    def __init__(self, target_metric, metric_reports: [ScoringMetricReport],
                 confusion_matrix: ConfusionMatrixReport = None, scatter: ScatterReport = None,
                 classification_report: SklearnClassificationReport = None):
        self._target_metric = target_metric
        self._metric_scores_dict = {r.metric_name: r for r in metric_reports}
        self._confusion_matrix = confusion_matrix
        self._scatter = scatter
        self._classification_report = classification_report

    def plot(self, figsize=(12,5)):
        if self.confusion_matrix is not None:
            self.confusion_matrix.plot(figsize=figsize)
        if self.scatter is not None:
            self.scatter.plot(figsize=figsize)
        if self.classification_report is not None:
            self.classification_report.plot(figsize=figsize)

        n = len(self.metric_scores)
        fig, axs = plt.subplots((n+1)//2, 2, squeeze=False, figsize=(12,n*2))
        for i, k in enumerate(sorted(self.metric_scores.keys())):
            self.metric_scores[k].plot(axs[i//2, i % 2])

        plt.tight_layout(pad=3)

    def to_dict(self, printable=False):
        metric_scores = {k: v.to_dict(printable=printable) for k, v in self.metric_scores.items()}
        for v in metric_scores.values():
            v.pop('metric', None)

        out = dict(
            metric_scores=metric_scores,
            target_metric=self.target_metric,
            scatter=None if self.scatter is None else self.scatter.to_dict(printable=printable),
            confusion_matrix=None if self.confusion_matrix is None else self.confusion_matrix.to_dict(
                printable=printable),
            classification_report=None if self.classification_report is None else self.classification_report.to_dict(
                printable=printable)
        )
        return self._printable_dict(out, printable=printable)

    @classmethod
    def to_dict_meaning(cls):
        return dict(
            target_metric="Metric of interest",
            metric_scores="Score information for various metrics saved in a dict structure where key is the metric name and value is of type {}".format(
                ScoringMetricReport.__name__),
            scatter="Scatter information (y_true vs y_pred). Available only for regressors.",
            confusion_matrix="Confusion matrix (y_true vs y_pred). Available only for classifiers.",
            classification_report="Sklearn's classification report"
        )

    def insights(self):
        insights_from_reports = [report.insights() for report in [self.metric_scores[self.target_metric],
                                                                  self.scatter, self.confusion_matrix,
                                                                  self.classification_report]
                                 if report is not None]

        return list(itertools.chain(*insights_from_reports))

        # metric_scores=metric_scores,
        # target_metric=self.target_metric,
        # scatter=None if self.scatter is None else self.scatter.to_dict(),
        # confusion_matrix=None if self.confusion_matrix is None else self.confusion_matrix.to_dict(),
        # classification_report=None if self.classification_report is None else self.classification_report.to_dict()


    @property
    def target_metric(self):
        return self._target_metric

    @property
    def metric_scores(self) -> dict:
        return self._metric_scores_dict

    @property
    def confusion_matrix(self) -> ConfusionMatrixReport:
        return self._confusion_matrix

    @property
    def scatter(self) -> ScatterReport:
        return self._scatter

    @property
    def classification_report(self) -> SklearnClassificationReport:
        return self._classification_report


if __name__ == '__main__':
    from pprint import pprint
    pprint(ScoringFullReport.to_dict_meaning(), width=160)


import unittest

import numpy
from sklearn.tree import DecisionTreeClassifier, DecisionTreeRegressor

from pytolemaic.analysis_logic.model_analysis.sensitivity.sensitivity import \
    SensitivityAnalysis
from pytolemaic.analysis_logic.model_analysis.sensitivity.sensitivity_reports import SensitivityTypes
from pytolemaic.utils.dmd import DMD
from pytolemaic.utils.general import GeneralUtils
from pytolemaic.utils.metrics import Metrics


class TestSensitivity(unittest.TestCase):

    def get_data(self, is_classification, seed=0):
        rs = numpy.random.RandomState(seed)
        x = rs.rand(10000, 10)
        x[:, 1] = 0
        # 1st is double importance, 2nd has no importance
        y = numpy.sum(x, axis=1) + 2 * x[:, 0]
        if is_classification:
            y = numpy.round(y, 0).astype(int)
        return DMD(x=x, y=y,
                   columns_meta={DMD.FEATURE_NAMES: ['f_' + str(k) for k in
                                                     range(x.shape[1])]})

    def get_model(self, is_classification):
        if is_classification:
            estimator = DecisionTreeClassifier
        else:
            estimator = DecisionTreeRegressor

        model = model = GeneralUtils.simple_imputation_pipeline(
            estimator(random_state=0))

        return model

    def test_sensitivity_raw_shuffled_classification(self,
                                                     is_classification=True):
        sensitivity = SensitivityAnalysis()
        model = self.get_model(is_classification)

        train = self.get_data(is_classification)
        model.fit(train.values, train.target.ravel())

        test = self.get_data(is_classification, seed=1)

        raw_scores = sensitivity.sensitivity_analysis(model=model,
                                                      metric=Metrics.recall.name,
                                                      dmd_test=test,
                                                      method=SensitivityTypes.shuffled,
                                                      raw_scores=True)

        raw_scores = raw_scores.sensitivities
        self.assertTrue(isinstance(raw_scores, dict))
        self.assertEqual(len(raw_scores), test.n_features)
        self.assertLessEqual(raw_scores['f_0'], 0.5)
        self.assertEqual(raw_scores['f_1'], 1.0)
        self.assertLessEqual(max([v for v in raw_scores.values()]), 1.0)

        scores = [v for k, v in raw_scores.items() if k not in ['f_0', 'f_1']]
        self.assertLessEqual(numpy.std(scores), 0.05)

    def test_sensitivity_raw_missing_regression(self,
                                                is_classification=False):
        sensitivity = SensitivityAnalysis()
        model = self.get_model(is_classification)

        train = self.get_data(is_classification)
        model.fit(train.values, train.target.ravel())

        test = self.get_data(is_classification, seed=1)

        raw_scores = sensitivity.sensitivity_analysis(model=model,
                                                      metric=Metrics.mae.name,
                                                      dmd_test=test,
                                                      method=SensitivityTypes.missing,
                                                      raw_scores=True)

        raw_scores = raw_scores.sensitivities
        self.assertTrue(isinstance(raw_scores, dict))
        self.assertEqual(len(raw_scores), test.n_features)
        self.assertEqual(raw_scores['f_1'], 0)
        self.assertGreaterEqual(raw_scores['f_0'], 0.8)
        self.assertGreaterEqual(min([v for v in raw_scores.values()]), 0)

        scores = [v for k, v in raw_scores.items() if k not in ['f_0', 'f_1']]
        self.assertLessEqual(numpy.std(scores), 0.05)

    def test_sensitivity_impact_regression(self, is_classification=False):
        sensitivity = SensitivityAnalysis()
        model = self.get_model(is_classification)

        train = self.get_data(is_classification)
        model.fit(train.values, train.target.ravel())

        test = self.get_data(is_classification, seed=1)

        scores = sensitivity.sensitivity_analysis(model=model,
                                                  metric=Metrics.mae.name,
                                                  dmd_test=test,
                                                  method=SensitivityTypes.missing,
                                                  raw_scores=False)

        scores = scores.sensitivities
        print(scores)
        self.assertTrue(isinstance(scores, dict))
        self.assertEqual(len(scores), test.n_features)
        self.assertGreaterEqual(numpy.round(sum([v for v in scores.values()]), 6), 1-1e-5)
        self.assertEqual(scores['f_1'], 0)
        self.assertGreaterEqual(scores['f_0'], 2 / len(scores))

        o_scores = [v for k, v in scores.items() if k not in ['f_0', 'f_1']]
        self.assertLessEqual(numpy.std(o_scores), 0.05)
        self.assertGreaterEqual(scores['f_0'], numpy.mean(o_scores))

    def test_sensitivity_impact_classification(self, is_classification=True):
        sensitivity = SensitivityAnalysis()
        model = self.get_model(is_classification)

        train = self.get_data(is_classification)
        model.fit(train.values, train.target.ravel())

        test = self.get_data(is_classification, seed=1)

        scores = sensitivity.sensitivity_analysis(model=model,
                                                  metric=Metrics.mae.name,
                                                  dmd_test=test,
                                                  method=SensitivityTypes.missing,
                                                  raw_scores=False)

        scores = scores.sensitivities
        print(scores)
        self.assertTrue(isinstance(scores, dict))
        self.assertEqual(len(scores), test.n_features)
        self.assertGreaterEqual(numpy.round(sum([v for v in scores.values()]), 6), 1-1e-5)
        self.assertEqual(scores['f_1'], 0)
        self.assertGreaterEqual(scores['f_0'], 2 / len(scores))

        o_scores = [v for k, v in scores.items() if k not in ['f_0', 'f_1']]
        self.assertLessEqual(numpy.std(o_scores), 0.05)
        self.assertGreaterEqual(scores['f_0'], numpy.mean(o_scores))

    def test_sensitivity_functions(self, is_classification=False):
        sensitivity = SensitivityAnalysis()
        model = self.get_model(is_classification)

        train = self.get_data(is_classification)
        model.fit(train.values, train.target.ravel())

        test = self.get_data(is_classification, seed=1)

        shuffled = sensitivity.sensitivity_analysis(model=model,
                                                   metric=Metrics.mae.name,
                                                   dmd_test=test,
                                                   method=SensitivityTypes.shuffled,
                                                   raw_scores=False)

        missing = sensitivity.sensitivity_analysis(model=model,
                                                   metric=Metrics.mae.name,
                                                   dmd_test=test,
                                                   method=SensitivityTypes.missing,
                                                   raw_scores=False)

        stats = sensitivity._sensitivity_stats_report(shuffled.sensitivities)
        n_features = stats.n_features
        n_zero = stats.n_zero
        n_very_low = stats.n_very_low
        n_low = stats.n_low


        leakage_score = sensitivity._leakage(
            n_features=n_features,
            n_very_low=n_very_low,
            n_zero=n_zero)

        self.assertGreater(leakage_score, 0)
        self.assertLessEqual(leakage_score, 1)

        overfit_score = sensitivity._too_many_features(
            n_features=n_features,
            n_very_low=n_very_low,
            n_low=n_low,
            n_zero=n_zero)

        self.assertGreaterEqual(overfit_score, 0)
        self.assertLessEqual(overfit_score, 1)

        imputation_score = sensitivity._imputation_score(shuffled=shuffled,
                                                         missing=missing)
        self.assertGreaterEqual(imputation_score, 0)
        self.assertLessEqual(imputation_score, 1)

        report = sensitivity._vulnerability_report(
            shuffled_sensitivity=shuffled,
            missing_sensitivity=missing)
        self.assertTrue(0 <= report.imputation <= 1)
        self.assertTrue(0 <= report.leakage <= 1)
        self.assertTrue(0 <= report.too_many_features <= 1)



import logging

import numpy
from sklearn import tree
from sklearn.base import ClassifierMixin
from sklearn.impute import SimpleImputer
from sklearn.tree import DecisionTreeClassifier, DecisionTreeRegressor, _tree

from pytolemaic.utils.constants import FeatureTypes
from pytolemaic.utils.dmd import DMD
from pytolemaic.utils.general import GeneralUtils
from pytolemaic.utils.metrics import Metrics
from resources.datasets.california_housing import CaliforniaHousing


class DecisionTreeExplainer():

    def __init__(self, n_samples=100000, explanation_depth=4, min_samples_leaf=100, allowed_delta=0.05, digitize=2,
                 cat_imputer=None, num_imputer=None):
        """

        :param n_samples: number of samples to create near the sample for decision tree model training
        :param explanation_depth: default depth of decision tree. Actual depth may be changed according to allowed_delta or actual path needed.
        :param min_samples_leaf: min_samples_leaf parameter for decision tree.
        :param allowed_delta: created decision tree must match the model's prediction up to this tolerance
        :param digitize: rounding level of numeric values for better explainability. Increase value if high resolution is required.
        :param cat_imputer: imputer for categorical features. By default SimpleImputer("most_frequent")
        :param num_imputer: imputer for numerical features. By default SimpleImputer("mean")

        """
        self.n_samples = n_samples
        self.explanation_depth = explanation_depth
        self.min_samples_leaf = min_samples_leaf
        self.is_classification = None
        self.dt = None
        self.train_data_stats = None
        self.model = None
        self.cache = None
        self.allowed_delta = allowed_delta
        self.digitize = digitize

        self.imputers = (
        cat_imputer or SimpleImputer(strategy='most_frequent'), num_imputer or SimpleImputer(strategy='mean'))

    def _create_dt(self, max_depth):
        self.is_classification = GeneralUtils.is_classification(self.model)
        if self.is_classification:
            dt = DecisionTreeClassifier(max_depth=max_depth, min_samples_leaf=self.min_samples_leaf, random_state=0,
                                        min_impurity_decrease=0.0001)
        else:
            dt = DecisionTreeRegressor(max_depth=max_depth, min_samples_leaf=self.min_samples_leaf, random_state=0,
                                       min_impurity_decrease=0.0001)

        return dt

    def _call_imputer(self, values, features, oper='fit', key=0):
        if oper == 'fit':
            if features is not None and len(features) > 0:
                self.imputers[key].fit(values[:, features])
        else:
            if features is not None and len(features) > 0:
                values[:, features] = self.imputers[key].transform(values[:, features])

    def fit(self, dmd_train: DMD, model):

        self.model = model
        self.train_data_stats = self._calc_train_data_stats(dmd_train)
        if dmd_train.categorical_features is not None:
            self._call_imputer(dmd_train.values, features=dmd_train.categorical_features, oper='fit', key=0)
            self._call_imputer(dmd_train.values, features=dmd_train.numerical_features, oper='fit', key=1)
        else:
            self._call_imputer(dmd_train.values, features=numpy.arange(dmd_train.n_features), oper='fit', key=1)

    def _calc_train_data_stats(self, dmd_train: DMD):
        return dmd_train  # TODO: save stats instead of entire train data

    def _create_neighborhood(self, sample: numpy.ndarray, n_samples, train_data_stats: DMD):
        random_state = numpy.random.RandomState(0)

        values = random_state.permutation(train_data_stats.values)
        mean, std = numpy.nanmean(values, axis=0), numpy.nanstd(values, axis=0)
        mn, mx = numpy.nanmin(values, axis=0), numpy.nanmax(values, axis=0)

        if train_data_stats.categorical_features is not None:
            self._call_imputer(values, features=train_data_stats.categorical_features, oper='transform', key=0)
            self._call_imputer(values, features=train_data_stats.numerical_features, oper='transform', key=1)
        else:
            self._call_imputer(values, features=numpy.arange(train_data_stats.n_features), oper='transform', key=1)
        # make sure we have n_Samples points
        while len(values) < n_samples:
            values = numpy.concatenate([values, values], axis=0)

        values = values[:n_samples, :]

        # create neighborhood
        for icol, feature_type in enumerate(train_data_stats.feature_types):

            if feature_type == FeatureTypes.categorical:
                # values[:, icol] = values[:, icol] == sample[icol]
                vec = random_state.permutation(values[:, icol])
                if not numpy.isnan(sample[icol]):
                    vec[:len(vec) // 2] = sample[icol]
                values[:, icol] = random_state.permutation(vec)
            else:
                # quantiles = numpy.percentile(values[:,icol], numpy.arange(101))
                # quantiles[0] = quantiles[0]-1e-5
                # quantiles[-1] = quantiles[-1]+1e-5
                # up, low = numpy.argwhere(quantiles>sample[icol])[0], numpy.argwhere(sample[icol]>quantiles)[-1]
                # up = quantiles[min(up+5, len(quantiles)-1)]
                # low = quantiles[max(0, low-5)]
                # values[values[:, icol]>=up, icol] = up
                # values[values[:, icol]<=low, icol] = low
                if not numpy.isnan(sample[icol]):
                    sigma = std[icol] / 10

                    values[:, icol] = sample[icol] + random_state.randn(len(values)) * sigma
                    values[:, icol] = numpy.clip(values[:, icol], mn[icol], mx[icol])

                else:
                    values[:, icol] = random_state.permutation(values[:, icol])
                values[:, icol] = numpy.round(values[:, icol], self.digitize) + 0.005

                # uniques = numpy.unique(values[:, icol])
                # len_uniques = len(uniques)
                # if len_uniques > 100:
                #     bins = numpy.linspace(mn[icol], mx[icol], num=100).tolist() + [mx[icol]+1]
                # else:
                #     bins = uniques.tolist() + [mx[icol]+1]
                # bins = numpy.array(bins)
                # inds = numpy.digitize(values[:, icol], bins=bins, right=False)
                # values[:, icol] = numpy.round(0.5*(bins[inds-1] + bins[inds]), 2)

        logging.info('size of neighborhood:', values.shape)
        return values

    @classmethod
    def _decision_path_to_text(cls, decision_tree, sample, feature_names):
        nodes = decision_tree.decision_path(sample.reshape(1, -1)).indices
        return cls._nodes_path_to_text(decision_tree, nodes_path=nodes, feature_names=feature_names)

    @classmethod
    def _nodes_path_to_text(cls, decision_tree, nodes_path, feature_names):

        nodes = nodes_path

        list_of_conditions = cls._list_of_conditions(decision_tree, feature_names, nodes)

        is_classification = isinstance(decision_tree, ClassifierMixin)
        if is_classification:
            value_msg = 'the classes probabilities are '
            values = decision_tree.tree_.value[nodes[-1]][0]

            values = values / numpy.sum(values)
            values = numpy.round(values, 3)
            value_msg += '{}'.format(values)
        else:
            mean = decision_tree.tree_.value[nodes[-1]][0][0]
            mean = numpy.round(mean, 3)
            mse = decision_tree.tree_.impurity[nodes[-1]]
            sigma = numpy.sqrt(mse)
            value_msg = 'the predicted value is {} +- {:.1g}'.format(mean, sigma)

        msg = 'Since ' + ' and '.join(list_of_conditions) + ' then {}'.format(value_msg)
        depth = len(nodes)
        return msg, depth

    @classmethod
    def _list_of_conditions(cls, decision_tree, feature_names, nodes):
        features = decision_tree.tree_.feature[nodes]
        names = [feature_names[i] for i in features if i >= 0]
        thresholds = decision_tree.tree_.threshold[nodes]
        left = decision_tree.tree_.children_left[nodes]
        list_of_questions = []  # give condition per node in path
        dict_of_conditions = {}  # condition per feature in path.
        feature_lower_limit = {}
        feature_upper_limit = {}
        features_involved = []
        # since names is ordered, the condition gets tighter along the way, no need to compare limit with previous one.
        for i in range(len(names)):
            name = names[i]
            threshold = GeneralUtils.f3(thresholds[i])
            if left[i] in nodes:
                question = '[{} <= {}]'.format(name, threshold)
                feature_upper_limit[names[i]] = threshold
            else:
                question = '[{} > {}]'.format(name, threshold)
                feature_lower_limit[names[i]] = threshold

            list_of_questions.append(question)

        keys = set(list(feature_lower_limit.keys()) + list(feature_upper_limit.keys()))
        for k in keys:
            c1 = '{} < '.format(feature_lower_limit[k]) if k in feature_lower_limit else ''
            c2 = '"{}"'.format(k)
            c3 = ' <= {}'.format(feature_upper_limit[k]) if k in feature_upper_limit else ''

            condition = '[{}{}{}]'.format(c1, c2, c3)
            dict_of_conditions[k] = condition

        list_of_conditions = []
        for f in feature_names:
            if f in dict_of_conditions:
                list_of_conditions.append(dict_of_conditions[f])

        return list_of_conditions

    def _fit_decision_tree(self, neiborhood, sample=None):
        x = neiborhood
        y = self.model.predict(x)

        delta = 1e100
        prev_delta = delta
        depth = self.explanation_depth
        while delta > self.allowed_delta:

            dt = self._create_dt(max_depth=depth)
            dt.fit(x, y)
            if self.is_classification:
                yt_proba = self.model.predict_proba(sample.reshape(1, -1))
                yp_proba = dt.predict_proba(sample.reshape(1, -1))
                delta = numpy.max(numpy.abs(yt_proba - yp_proba))
            else:
                yt = self.model.predict(sample.reshape(1, -1))
                yp = dt.predict(sample.reshape(1, -1))
                delta = 2 * numpy.abs(yt - yp) / (numpy.abs(yt + yp) + 1e-3)

            depth += 1
            if delta == prev_delta:
                raise ValueError("Failed training surrogate dt: dt does not converge correctly on given sample.\n"
                                 "Either increase allowed_delta (>{}) or decrease min_samples_leaf".format(delta))
            else:
                prev_delta = delta

        msg, depth = self._decision_path_to_text(decision_tree=dt, sample=sample,
                                                 feature_names=self.train_data_stats.feature_names)

        # check dt performance
        n = len(x) // 5
        train, valid = x[n:, :], x[:n, :]

        dt = self._create_dt(max_depth=depth)
        dt.fit(train, y[n:])

        if self.is_classification:
            logging.info('recall score=', Metrics.recall.function(y[:n], dt.predict(valid)))
            # logging.info("model prediction = {}, surrogate dt model = {}".format(self.model.predict_proba(sample.reshape(1,-1)), dt.predict_proba(sample.reshape(1,-1))))

        else:
            logging.info('r2 score=', Metrics.r2.function(y[:n], dt.predict(valid)))
            # logging.info("model prediction = {}, surrogate dt model = {}".format(self.model.predict(sample.reshape(1,-1)), dt.predict(sample.reshape(1,-1))))

        # train dt on entire data
        dt.fit(x, y)

        return dt

    def _fit_dt_on_neighborhood(self, sample):
        if self.cache is None or self.cache != str(sample):
            self.cache = str(sample)
            self.dt = self._fit_decision_tree(self._create_neighborhood(sample=sample, n_samples=self.n_samples,
                                                                        train_data_stats=self.train_data_stats),
                                              sample=sample)
        else:
            pass  # dt is already fitted

    def explain(self, sample: numpy.ndarray):
        self._fit_dt_on_neighborhood(sample=sample)
        msg, depth = self._decision_path_to_text(decision_tree=self.dt, sample=sample,
                                                 feature_names=self.train_data_stats.feature_names)
        return msg

    def plot(self, sample: numpy.ndarray):
        self._fit_dt_on_neighborhood(sample=sample)
        tree.plot_tree(self.dt, feature_names=self.train_data_stats.feature_names, fontsize=10, filled=True)

        try:
            from matplotlib import pyplot as plt

            msg = self.explain(sample)
            msg = msg.replace('and', 'and\n').replace('then', '\nthen')
            plt.text(.0, 1., msg,
                     transform=plt.gca().transAxes, size=12,
                     horizontalalignment='left',
                     verticalalignment='top')
        except:
            raise


if __name__ == '__main__':
    dataset = CaliforniaHousing()
    model = dataset.get_model()
    from matplotlib import pyplot as plt

    plt.rcParams['figure.figsize'] = (20, 10)

    train = dataset.as_dmd()[0]
    sample = dataset.testing_data[0][3, :]
    explainer = DecisionTreeExplainer()
    explainer.fit(train, model)
    logging.info('\n'.join([train.feature_names[icol] + ' : ' + str(sample[icol]) for icol in range(train.n_features)]))

    logging.info('\n\nexplain')
    explainer.explain(sample)
    logging.info('\n\nplot')
    explainer.plot(sample)

    plt.show()



import numpy
import sklearn.datasets
import sklearn.model_selection
from matplotlib import pyplot as plt
from sklearn.tree import DecisionTreeClassifier

from pytolemaic import PyTrust


def run():
    # Dataset: xtrain, ytrain, xtest, ytest
    # noinspection PyUnresolvedReferences
    data = sklearn.datasets.load_wine(return_X_y=False)

    x = data['data']
    y = data['target']
    feature_names = data['feature_names']
    labels = data['target_names']

    train_inds, test_inds = sklearn.model_selection.train_test_split(
        numpy.arange(len(data['data'])), test_size=0.3)

    xtrain, ytrain = x[train_inds], y[train_inds]
    xtest, ytest = x[test_inds], y[test_inds]

    # Train estimator
    estimator = DecisionTreeClassifier()
    estimator.fit(xtrain, ytrain)

    # Initiating PyTrust
    pytrust = PyTrust(
        model=estimator,
        xtrain=xtrain, ytrain=ytrain,
        xtest=xtest, ytest=ytest,
        metric='recall')

    # Initiating PyTrust with more information
    pytrust = PyTrust(
        model=estimator,
        xtrain=xtrain, ytrain=ytrain,
        xtest=xtest, ytest=ytest,
        feature_names=feature_names,
        target_labels={i: label for i, label in enumerate(labels)},
        splitter='stratified',
        metric='recall')

    pytrust.scoring_report.plot()

    pytrust.sensitivity_report.plot()

    pytrust.dataset_analysis_report.plot()

    pytrust.quality_report.plot()

    sample = xtest[0, :].reshape(1, -1)
    explainer = pytrust.create_lime_explainer(max_samples=16000)
    explainer.explain(sample=sample)

    uncertainty_model = pytrust.create_uncertainty_model(method='default')
    prediction = uncertainty_model.predict(sample)  # same as model.predict
    uncertainty = uncertainty_model.uncertainty(sample)  # uncertainty value

    print("Let's check for insights...")
    print('\n'.join(pytrust.insights))
    print("Done!")


if __name__ == '__main__':
    run()
    plt.show()



import numpy
from matplotlib import pyplot as plt

from pytolemaic import Metrics
from pytolemaic import PyTrust
from pytolemaic.utils.general import GeneralUtils
from resources.datasets.california_housing import CaliforniaHousing


def run():
    # init
    dataset = CaliforniaHousing()
    regressor = dataset.get_model()
    train, test = dataset.as_dmd()

    test_1st_half, test_2nd_half = test.split(ratio=0.2)

    metric = Metrics.mae

    pytrust = PyTrust(
        model=regressor,
        xtest=test_1st_half,
        metric=metric)

    xtest2, ytest2 = test_2nd_half.values, test_2nd_half.target

    method = 'mae'  # or 'probability'
    uncertainty_model = pytrust.create_uncertainty_model(method=method)
    yp = uncertainty_model.predict(xtest2)  # same as model.predict
    uncertainty = uncertainty_model.uncertainty(xtest2)  # uncertainty value
    print('y_true, y_pred, uncertainty')
    print(numpy.concatenate([ytest2.reshape(-1, 1), yp.reshape(-1, 1), uncertainty.reshape(-1, 1)], axis=1)[:10])

    # example
    plt.figure()
    uncertainty_levels = numpy.array([0, 0.2, 0.4, 0.6, 0.8, 1.0001])
    mn, mx = 1, 0

    # uncertainty model may be based on 'confidence' or 'probability' for classification, and 'mae' or 'rmse' for regression
    for method in ['quantile', 'mae', 'rmse']:
        print('working on method %s' % method)
        # train uncertainty model
        uncertainty_model = pytrust.create_uncertainty_model(method=method)
        yp = uncertainty_model.predict(xtest2)  # same as model.predict
        uncertainty = uncertainty_model.uncertainty(xtest2)  # uncertainty value

        level_inds = numpy.digitize(uncertainty.ravel(), uncertainty_levels)

        performance = []
        for ibin in range(len(uncertainty_levels) - 1):
            inds = level_inds == ibin + 1
            if not any(inds):
                performance.append(0)
            else:
                subset_score = metric.function(y_true=ytest2[inds], y_pred=yp[inds])
                performance.append(subset_score)

        uncertainty_levels_middle = (uncertainty_levels[1:] + uncertainty_levels[:-1]) / 2

        plt.figure(1)
        plt.plot(uncertainty_levels_middle, performance, '*-b' if method == 'mae' else '*-r')

        plt.xlabel("Uncertainty level")
        plt.ylabel("{} Score".format(metric.name))
        plt.title("{} score vs uncertainty level".format(metric.name))
        plt.legend(['method=mae', 'method=rmse', 'method=quantile'], loc='upper right')

        print(uncertainty_levels_middle)
        print(GeneralUtils.f3(performance))
        mn = min(min(performance), mn)
        mx = max(max(performance), mx)
        uncertainty_model.plot_calibration_curve()

    plt.figure(1)
    # emphasize bins
    for level in uncertainty_levels:
        plt.plot([level, level], [mn, mx], '-k')


if __name__ == '__main__':
    run()
    plt.show()



import itertools

from matplotlib import pyplot as plt

from pytolemaic.utils.base_report import Report

class SensitivityTypes():
    shuffled = 'shuffled'
    missing = 'missing'


class SensitivityStatsReport(Report):
    def __init__(self, n_features: int, n_low: int, n_very_low: int, n_zero: int):
        self._n_features = n_features
        self._n_low = n_low
        self._n_very_low = n_very_low
        self._n_zero = n_zero

    def to_dict(self, printable=False):
        out = dict(
            n_features=self.n_features,
            n_low=self.n_low,
            n_very_low=self.n_very_low,
            n_zero=self.n_zero,
        )
        return self._printable_dict(out, printable=printable)

    @classmethod
    def to_dict_meaning(cls):
        return dict(
            n_features="Number of features in dataset",
            n_low="Number of feature with low sensitivity (sensitivity lower than 5% of max sensitivity)",
            n_very_low="Number of feature with low sensitivity (sensitivity lower than 5% of max sensitivity)",
            n_zero="Number of feature with zero sensitivity",
        )

    def plot(self, ax=None, method=None):
        if ax is None:
            fig, ax = plt.subplots(1)

        keys, values = zip(*sorted(self.to_dict().items()))

        title_prefix = '"{}" '.format(method) if method else ''
        ax.bar(keys, values)
        ax.set(
            title='{}Sensitivity Statistics'.format(title_prefix),
            ylabel='# of features')
        plt.draw()
        plt.tight_layout()

    @property
    def n_features(self):
        return self._n_features

    @property
    def n_low(self):
        return self._n_low

    @property
    def n_very_low(self):
        return self._n_very_low

    @property
    def n_zero(self):
        return self._n_zero

    def insights(self):

        insights = []

        lvl = 0.5
        sentence = "More than {} of features have".format(int(lvl * self.n_features))
        if self.n_zero > lvl * self.n_features:
            insights.append("{} no sensitivity at all".format(sentence))
        elif self.n_very_low > lvl * self.n_features:
            insights.append("{} very low sensitivity".format(sentence))
        elif self.n_low > lvl * self.n_features:
            insights.append("{} low sensitivity".format(sentence))

        # insights = self._add_cls_name_prefix(insights)

        return insights


class SensitivityVulnerabilityReport(Report):
    def __init__(self, imputation: float, leakage: float, too_many_features: float):
        self._too_many_features = too_many_features
        self._imputation = imputation
        self._leakage = leakage

    def to_dict(self, printable=False):
        out = dict(
            too_many_features=self.too_many_features,
            imputation=self.imputation,
            leakage=self.leakage
        )
        return self._printable_dict(out, printable=printable)

    @classmethod
    def to_dict_meaning(cls):
        return dict(
            too_many_features="Using many features slow down model's training, affect data pipeline and preprocessing and may also make model more susceptible to overfit. Higher value means there are many redundant features in dataset.",
            imputation="Sensitivity of the features should be similar in all methods. If not, this may indicate an issue with the imputation process. Higher value means the imputation process has higher impact on the model.",
            leakage="Measure how likely the model is prone to data leakage."
        )

    def plot(self, ax=None):
        if ax is None:
            fig, ax = plt.subplots(1)

        keys, values = zip(*sorted(self.to_dict().items()))

        ax.bar(keys, values)
        ax.set(
            title="Model's vulnerability metrics",
            ylabel='Vulnerability scores',
            ylim=[0, 1])
        plt.draw()

    @property
    def imputation(self):
        return self._imputation

    @property
    def leakage(self):
        return self._leakage

    @property
    def too_many_features(self):
        return self._too_many_features

    def insights(self):

        insights = []
        lvl1, lvl2, lvl3 = 0.1, 0.5, 0.9
        sentence = "Vulnerability to imputation is {:.3g} indicating your model is %s to the imputation technique.".format(
            self.imputation)
        if self.imputation < lvl1:
            pass  # ok
        elif self.imputation < lvl2:
            insights.append(sentence % "somewhat sensitive")
        elif self.imputation < lvl3:
            insights.append(sentence % "sensitive")
        else:
            insights.append(sentence % "very sensitive" + "You should check what's going on, there may be a bug.")

        lvl1, lvl2 = 0.1, 0.75
        sentence = "Vulnerability to data leakage is {:.3g} indicating %s data leakage.".format(self.leakage)
        if self.leakage < lvl1:
            pass  # ok
        elif self.leakage < lvl2:
            insights.append(
                sentence % "there is a chance of" + " Check the features with the highest sensitivity scores.")
        else:
            insights.append(
                sentence % "a very high chance of" + " Check the features with the highest sensitivity scores.")

        lvl1, lvl2 = 0.1, 0.5
        sentence = "Vulnerability to number of features is {:.3g} indicating %s.".format(self.too_many_features)
        if self.too_many_features < lvl1:
            pass  # ok
        elif self.too_many_features < lvl2:
            insights.append(
                sentence % "there are substantial number of features which could be discarded / regularized")
        else:
            insights.append(
                sentence % "that many features have little value and may cause overfit. Discard these feature or increase regularization.")

        return self._add_cls_name_prefix(insights)

class SensitivityOfFeaturesReport(Report):
    def __init__(self, method: str, sensitivities: dict,
                 stats_report: SensitivityStatsReport):
        self._method = method
        self._sensitivities = sensitivities
        self._stats_report = stats_report

    def to_dict(self, printable=False):
        out = dict(
            method=self.method,
            sensitivities=self.sensitivities,
            stats=self.stats_report.to_dict(printable=printable)
        )
        return self._printable_dict(out, printable=printable)

    @classmethod
    def to_dict_meaning(cls):
        return dict(
            method="Method used to calculate sensitivity",
            sensitivities="key-value dictionary where the key is feature name and value is feature sensitivity",
            stats=SensitivityStatsReport.to_dict_meaning()
        )

    def plot_sorted_sensitivities(self, ax=None, n_features_to_plot=10):
        if ax is None:
            fig, ax = plt.subplots(1, 1, figsize=(10, 10))

        sorted_features = self.sorted_sensitivities  # sorted(self.sensitivities.items(), key=lambda kv: -kv[1])

        if n_features_to_plot is not None:
            sorted_features = sorted_features[:min(n_features_to_plot, len(sorted_features))]

        keys, values = zip(*sorted_features)
        keys = ['"{}"'.format(k) for k in keys]
        ax.barh(list(reversed(keys)), list(reversed(values)))
        ax.set(
            title='"{}" Feature Sensitivity'.format(self.method),
            xlabel='Sensitivity value')

    def plot(self, axs=None, n_features_to_plot=10):
        if axs is None:
            fig, axs = plt.subplots(1, 2, figsize=(10, 10))

        ax1, ax2 = axs
        self.plot_sorted_sensitivities(ax=ax1, n_features_to_plot=n_features_to_plot)
        self.stats_report.plot(ax=ax2, method=self.method)
        plt.draw()

    @property
    def method(self):
        return self._method

    @property
    def sensitivities(self):
        return self._sensitivities

    @property
    def sorted_sensitivities(self):
        return [(k, v) for k, v in sorted(self._sensitivities.items(), key=lambda kv: -kv[1])]

    @property
    def stats_report(self) -> SensitivityStatsReport:
        return self._stats_report

    def most_important_features(self, n_features=3):
        out = []
        n_features = min(n_features, len(self.sorted_sensitivities))
        for i in range(n_features):
            if self.sorted_sensitivities[i][1] > 0:
                out.append(self.sorted_sensitivities[i][0])
        return out

    def most_important_features_insight(self):
        features = self.most_important_features(n_features=3)
        msg = ''
        if len(features)>=1:
            msg += "The most important feature is '{}'".format(features[0])
        if len(features)>=2:
            msg += ", followed by '{}'".format(features[1])
        if len(features) >= 3:
            msg += "and '{}'".format(features[2])

        return msg+"."

    def insights(self):
        stats_report_insights = self.stats_report.insights()
        insights = []
        insights.append(self.most_important_features_insight())
        if self.stats_report.n_zero > 0:
            zero_sensitivity = [feature for feature, value in self.sorted_sensitivities[-self.stats_report.n_zero:]]
            if len(zero_sensitivity) == 1:
                list_of_features = " '{}'".format(zero_sensitivity[0])
            else:
                list_of_features = "\n\t" + ", ".join(["'{}'".format(feature) for feature in zero_sensitivity])

            insights.append("The following features has 0 sensitivity:{}".format(list_of_features))

        if self.stats_report.n_very_low - self.stats_report.n_zero > 0:
            very_low_sensitivity = [feature for feature, value in
                                    self.sorted_sensitivities[
                                    -self.stats_report.n_very_low - self.stats_report.n_zero: -self.stats_report.n_zero]]
            if len(very_low_sensitivity) == 1:
                list_of_features = " '{}'".format(very_low_sensitivity[0])
            else:
                list_of_features = "\n\t" + ", ".join(["'{}'".format(feature) for feature in very_low_sensitivity])

            insights.append(
                "The following features can be discarded due to very low sensitivity:{}".format(list_of_features))

        insights = itertools.chain(stats_report_insights, insights)
        insights = ["{}.{}: {}".format(type(self).__name__, self.method, insight) for insight in insights]
        return insights

class SensitivityFullReport(Report):

    def __init__(self,
                 shuffle_report: SensitivityOfFeaturesReport,
                 missing_report: SensitivityOfFeaturesReport,
                 vulnerability_report: SensitivityVulnerabilityReport
                 ):
        self._shuffle_report = shuffle_report
        self._missing_report = missing_report
        self._vulnerability_report = vulnerability_report

    def plot(self):
        if self.missing_report is not None:
            fig, (axs1, axs2) = plt.subplots(2, 2, figsize=(10, 10))
            self.shuffle_report.plot(axs=axs1)
            self.missing_report.plot(axs=axs2)

        else:
            fig, axs = plt.subplots(1, 2, figsize=(10, 10))
            self.shuffle_report.plot(axs=axs)

        plt.tight_layout()

        self.vulnerability_report.plot()

    @classmethod
    def to_dict_meaning(cls):
        return dict(
            shuffle_report=SensitivityOfFeaturesReport.to_dict_meaning(),
            missing_report=SensitivityOfFeaturesReport.to_dict_meaning(),
            vulnerability_report=SensitivityVulnerabilityReport.to_dict_meaning(),
        )

    def to_dict(self, printable=False):
        out = dict(
            shuffle_report=self.shuffle_report.to_dict(),
            missing_report=None if self.missing_report is None else self.missing_report.to_dict(),
            vulnerability_report=self.vulnerability_report.to_dict(),
        )
        return self._printable_dict(out, printable=printable)

    @property
    def shuffle_report(self) -> SensitivityOfFeaturesReport:
        return self._shuffle_report

    @property
    def missing_report(self) -> SensitivityOfFeaturesReport:
        return self._missing_report

    @property
    def vulnerability_report(self) -> SensitivityVulnerabilityReport:
        return self._vulnerability_report

    def insights(self):
        return list(itertools.chain(self.shuffle_report.insights(),
                                    [] if self.missing_report is None else self.missing_report.insights(),
                                    self.vulnerability_report.insights()))


if __name__ == '__main__':
    from pprint import pprint

    pprint(SensitivityFullReport.to_dict_meaning(), width=160)
